#!/usr/bin/env python3

"""
Pegasus utility functions for pasing a kickstart output file and return wanted information

"""

import logging
import re
import sys
import traceback
from pprint import pprint
from xml.parsers import expat

import yaml
import yaml.constructor

from Pegasus.monitoring.metadata import FileMetadata

##
#  Copyright 2007-2010 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##

# Revision : $Revision: 2012 $

yaml.constructor.SafeConstructor.yaml_constructors[
    "tag:yaml.org,2002:timestamp"
] = yaml.constructor.SafeConstructor.yaml_constructors["tag:yaml.org,2002:str"]

# Regular expressions used in the kickstart parser
re_parse_props = re.compile(r'(\S+)\s*=\s*([^",]+)')
re_parse_quoted_props = re.compile(r'(\S+)\s*=\s*"([^"]+)"')

logger = logging.getLogger(__name__)


class Parser:
    """
    This class is used to parse a kickstart output file, and return
    requested information.
    """

    def __init__(self, filename):
        """
        This function initializes the Parser class with the kickstart
        output file that should be parsed.
        """
        self._kickstart_output_file = filename
        self._ks_elements = {}
        self._keys = {}
        self._fh = None
        self._open_error = False
        self._parser = None

    def open(self):
        """
        This function opens a kickstart output file.
        """
        try:
            self._fh = open(self._kickstart_output_file)
        except Exception:
            # Error opening file
            self._fh = None
            self._open_error = True
            return False

        # Open succeeded
        self._open_error = False
        return True

    def close(self):
        """
        This function closes the kickstart output file.
        """
        try:
            self._fh.close()
        except Exception:
            return False

        return True

    def is_invocation_record(self, buffer=""):
        """
        Returns True if buffer contains an invocation record either xml or invocation
        """
        # first check for yaml
        if buffer.find("- invocation:") == -1:
            # no yaml check for xml
            if buffer.find("<invocation") == -1:
                return False
        return True

    def is_multipart_record(self, buffer=""):
        """
        Returns True if buffer contains an invocation record either xml or invocation
        """
        if self.is_invocation_record(buffer):
            return False
        else:
            return buffer.find("- ") == 0

    def is_task_record(self, buffer=""):
        """
        Returns True if buffer contains a task record.
        """
        if buffer.find("[seqexec-task") != -1 or buffer.find("[cluster-task") != -1:
            return True
        return False

    def is_clustered_record(self, buffer=""):
        """
        Returns True if buffer contains a clustered record.
        """
        if (
            buffer.find("[seqexec-summary") != -1
            or buffer.find("[cluster-summary") != -1
        ):
            return True
        return False

    def parse_clustered_record(self, buffer=""):
        """
        Parses the clustered record in buffer, returning all found keys
        """
        self._keys = {}

        # Check if we have an invocation record
        if self.is_clustered_record(buffer) is False:
            return self._keys

        # Add clustered key to our response
        self._keys["clustered"] = True

        # Parse all quoted properties
        for my_key, my_val in re_parse_quoted_props.findall(buffer):
            self._keys[my_key] = my_val

        # And add unquoted properties as well
        for my_key, my_val in re_parse_props.findall(buffer):
            self._keys[my_key] = my_val

        return self._keys

    def parse_task_record(self, buffer=""):
        """
        Parses the task record in buffer, returning all found keys
        """
        self._keys = {}

        # Check if we have an invocation record
        if self.is_task_record(buffer) is False:
            return self._keys

        # Add task key to our response
        self._keys["task"] = True

        # Parse all quoted properties
        for my_key, my_val in re_parse_quoted_props.findall(buffer):
            self._keys[my_key] = my_val

        # And add unquoted properties as well
        for my_key, my_val in re_parse_props.findall(buffer):
            self._keys[my_key] = my_val

        return self._keys

    def parse(self, keys_dict, tasks=True, clustered=True):
        """
        This function parses the kickstart output file, looking for
        the keys specified in the keys_dict variable. It returns a
        list of dictionaries containing the found keys. Look at the
        parse_stampede function for details about how to pass keys
        using the keys_dict structure. The function will return an
        empty list if no records are found or if an error happens.
        """

        # Place keys_dict in the _ks_elements
        self._ks_elements = keys_dict
        my_reply = []

        # Try to open the file
        if self.open() is False:
            return my_reply

        logger.debug(
            "Started reading records from kickstart file %s"
            % (self._kickstart_output_file)
        )

        # load the appropriate parser by checking for first instance of xml header or invocation
        yaml_parser = True
        for line in self._fh:
            if line.find("<?xml") != -1:
                yaml_parser = False
                break
            elif line.find("- invocation:") != -1:
                yaml_parser = True
                break

        self.close()
        self._parser = (
            YAMLParser(self._kickstart_output_file)
            if yaml_parser
            else XMLParser(self._kickstart_output_file)
        )

        return self._parser.parse(keys_dict, tasks, clustered)

    def parse_stampede(self):
        """
        This function works similarly to the parse function above,
        but does not require a keys_dict parameter as it uses a
        built-in list of keys speficically used in the Stampede
        schema.
        """

        stampede_elements = {
            "invocation": [
                "hostname",
                "resource",
                "user",
                "hostaddr",
                "transformation",
                "derivation",
            ],
            "mainjob": ["duration", "start"],
            "usage": ["utime", "stime", "maxrss"],
            "ram": ["total"],
            "uname": ["system", "release", "machine"],
            "file": ["name"],
            "status": ["raw"],
            "signalled": [
                "signal",
                "corefile",
                "action",
            ],  # action is the char data in signalled element
            "regular": ["exitcode"],
            "argument-vector": [],
            "cwd": [],
            "stdout": [],
            "stderr": [],
            "statinfo": ["lfn", "size", "ctime", "user"],
            "checksum": ["type", "value", "timing"],
            "type": ["type", "value"],
            "cpu": ["count", "speed", "vendor"],
        }

        return self.parse(stampede_elements, tasks=True, clustered=True)

    def parse_stdout_stderr(self):
        """
        This function extracts the stdout and stderr from a kickstart output file.
        It returns an array containing the output for each task in a job.
        """

        stdout_stderr_elements = {
            "invocation": ["hostname", "resource", "derivation", "transformation"],
            "file": ["name"],
            "regular": ["exitcode"],
            "failure": ["error"],
            "argument-vector": [],
            "cwd": [],
            "stdout": [],
            "stderr": [],
        }

        return self.parse(stdout_stderr_elements, tasks=False, clustered=False)


class YAMLParser(Parser):
    """
    Represents the parser that parses the kickstart xml records
    """

    def __init__(self, filename):
        super().__init__(filename)

    def parse(self, keys_dict, tasks=True, clustered=True):
        """
        This function parses the kickstart output file, looking for
        the keys specified in the keys_dict variable. It returns a
        list of dictionaries containing the found keys. Look at the
        parse_stampede function for details about how to pass keys
        using the keys_dict structure. The function will return an
        empty list if no records are found or if an error happens.
        """

        my_reply = []

        # Place keys_dict in the _ks_elements
        self._ks_elements = keys_dict

        # Try to open the file
        if self.open() is False:
            return my_reply

        logger.debug(
            "Started reading records from kickstart file %s"
            % (self._kickstart_output_file)
        )

        self._record_number = 0
        # Read first record
        record = self.read_record()

        # Loop while we still have record to read
        while record is not None:
            logger.debug("Record is \n%s" % record)
            if self.is_invocation_record(record) is True:
                # We have an invocation record, parse it!
                try:
                    my_record = self.parse_invocation_record(record)
                except Exception:
                    logger.warning(
                        "KICKSTART-PARSE-ERROR --> error parsing YAML invocation record in file %s"
                        % (self._kickstart_output_file)
                    )
                    logger.warning(traceback.format_exc())
                    # Found error parsing this file, return empty reply
                    my_reply = []
                    # Finish the loop
                    break
                my_reply.append(my_record)
            elif self.is_clustered_record(record) is True:
                # Check if we want clustered records too
                if clustered:
                    # Clustered records are seqexec summary records for clustered jobs
                    # We have a clustered record, parse it!
                    my_reply.append(self.parse_clustered_record(record))
            elif self.is_task_record(record) is True:
                # Check if we want task records too
                if tasks:
                    # We have a clustered record, parse it!
                    my_reply.append(self.parse_task_record(record))
            elif self.is_multipart_record(record) is True:
                logger.debug(
                    "Multipart Record in file %s" % (self._kickstart_output_file)
                )
                # can return multiple yaml snippets
                my_records = self.parse_multipart_record(record)
                for record in my_records:
                    my_reply.append(record)
            else:
                # We have something else, this shouldn't happen!
                # Just skip it
                pass

            # Read next record
            record = self.read_record()

        # Lastly, close the file
        self.close()

        return my_reply

    def read_record(self):
        """
        This function reads an invocation record from the kickstart
        output file. We also look for the struct at the end of a file
        containing multiple records. It returns a string containing
        the record, or None if it is not found.
        """
        buffer = ""

        # valid token that is parsed
        token = ""

        self._record_number += 1
        logger.debug(
            "Started reading record number %d from kickstart file %s"
            % (self._record_number, self._kickstart_output_file)
        )

        # First, we find the beginning <invocation xmlns....
        while True:
            line = self._fh.readline()
            if line == "":
                # End of file, record not found
                return None
            if line.find("- invocation:") != -1:
                token = "- invocation:"
                break
            if line.find("[cluster-task") != -1:
                token = "[cluster-task"
                break
            if line.find("[cluster-summary") != -1:
                token = "[cluster-summary"
                break
            if line.find("[seqexec-task") != -1:
                # deprecated token
                token = "[seqexec-task"
                break
            if line.find("[seqexec-summary") != -1:
                # deprecated token
                token = "[seqexec-summary"
                break
            if line.find("---------------pegasus-multipart") == 0:
                # token
                token = "pegasus-multipart"
                break

        # Found something!
        if token == "- invocation:":
            # Found invocation record
            start = line.find("- invocation:")
            buffer = line[start:]
            # Check if we have everything in a single line
            # Not clear what to do for that for YAML records
        elif token == "pegasus-multipart":
            buffer = ""
        elif token == "[cluster-summary" or token == "[seqexec-summary":
            # Found line with cluster jobs summary
            start = line.find(token)
            buffer = line[start:]
            end = buffer.find("]")

            if end >= 0:
                end = end + len("]")
                return buffer[:end]

            # clustered record should be in a single line!
            logger.warning(
                "%s: %s line is malformed... ignoring it..."
                % (self._kickstart_output_file, token)
            )
            return ""
        elif token == "[cluster-task" or token == "[seqexec-task":
            # Found line with task information
            start = line.find(token)
            buffer = line[start:]
            end = buffer.find("]")

            if end >= 0:
                end = end + len("]")
                return buffer[:end]

            # task record should be in a single line!
            logger.warning(
                "%s: %s line is malformed... ignoring it..."
                % (self._kickstart_output_file, token)
            )
            return ""
        else:
            return ""

        # Ok, now continue reading the file until we get a full record
        buffer = [buffer]

        while True:
            file_ptr = self._fh.tell()
            line = self._fh.readline()
            if line == "":
                # End of file
                break

            if (
                line.find("[cluster-") == 0
                or line.find("---------------pegasus-multipart") == 0
            ):
                # this is to trigger end of parsing of a single kickstart record
                logger.debug(
                    "Hit end of invocation record in file %s: "
                    % self._kickstart_output_file
                )
                # back track file pointer
                self._fh.seek(file_ptr)
                break
            elif line[0] in [" ", "-", "\n"]:
                buffer.append(line)

        record = "".join(buffer)
        logger.debug(
            "Finished reading record number %d from kickstart file %s"
            % (self._record_number, self._kickstart_output_file)
        )
        return record

    def dicts_remap(self, src, src_keys, dst, dst_keys):
        """
        Pulls data from a provided location in a src dict, and inserts
        the data at a provided location in the dst dic - this is used
        to transition from the old xml format to the new yaml format
        """
        for key in src_keys:
            if key in src:
                src = src[key]
            else:
                src = None
                break

        if src is None:
            return

        for key in dst_keys[:-1]:
            if key not in dst:
                dst[key] = {}
            dst = dst[key]

        dst[dst_keys[-1]] = src

    def map_yaml_to_ver2_format(self, data):
        """
        Maps from new yaml dict format to old v2 format we used with the xml records
        """
        # unmappable:
        #  "file": ["name"]

        # new format -> old format
        my_map = [
            [["hostname"], ["hostname"]],
            [["resource"], ["resource"]],
            [["user"], ["user"]],
            [["hostaddr"], ["hostaddr"]],
            [["transformation"], ["transformation"]],
            [["derivation"], ["derivation"]],
            [["mainjob", "duration"], ["duration"]],
            [["mainjob", "start"], ["start"]],
            [["mainjob", "usage", "utime"], ["utime"]],
            [["mainjob", "usage", "stime"], ["stime"]],
            [["mainjob", "usage", "maxrss"], ["maxrss"]],
            [["machine", "ram_total"], ["ram"]],
            [["machine", "uname_system"], ["system"]],
            [["machine", "uname_release"], ["release"]],
            [["machine", "uname_machine"], ["machine"]],
            [["mainjob", "executable", "file_name"], ["name"]],
            [["mainjob", "status", "raw"], ["raw"]],
            [["mainjob", "status", "signalled_signal"], ["signal"]],
            [["mainjob", "status", "signalled_name"], ["action"]],
            [["mainjob", "status", "corefile"], ["corefile"]],
            [["mainjob", "status", "regular_exitcode"], ["exitcode"]],
            [["cwd"], ["cwd"]],
            [["files", "stdout", "data"], ["stdout"]],
            [["files", "stderr", "data"], ["stderr"]],
        ]

        #        stampede_elements = {"invocation": ["hostname", "resource", "user", "hostaddr", "transformation", "derivation"],
        #                             "mainjob": ["duration", "start"],
        #                             "usage": ["utime", "stime"],
        #                             "ram": ["total"],
        #                             "uname": ["system", "release", "machine"],
        #                             "file": ["name"],
        #                             "status": ["raw"],
        #                             "signalled": ["signal", "corefile", "action"], #action is the char data in signalled element
        #                             "regular": ["exitcode"],
        #                             "argument-vector": [],
        #                             "cwd": [],
        #                             "stdout": [],
        #                             "stderr": [],
        #                             "statinfo": ["lfn", "size", "ctime", "user" ],
        #                             "checksum": ["type", "value", "timing"],
        #                             "type": ["type", "value"]}

        new_data = {}
        new_data["invocation"] = True
        new_data["checksum"] = {}
        new_data["outputs"] = {}
        for mapping in my_map:
            self.dicts_remap(data, mapping[0], new_data, mapping[1])

        # some mappings are based on lfns
        if "files" in data:
            for lfn in data["files"]:
                file_data = data["files"][lfn]
                output = file_data["output"] if "output" in file_data.keys() else False
                if not output:
                    continue
                meta = FileMetadata()
                meta._id = lfn

                """
                add whatever 4.9 attributes are
                  {
                    "_type": "file",
                    "_id": "f.b2",
                    "_attributes": {
                      "ctime": "2019-02-19T16:42:52-08:00",
                      "checksum.timing": "0.144",
                      "user": "vahi",
                      "checksum.type": "sha256",
                      "checksum.value": "4a77bee20a28a446506ef7531ffc038053f52e5211d93a95fe5193746af8d23a",
                      "size": "123"
                    }
                  },
                """
                if "user" in data["files"][lfn]:
                    meta.add_attribute("user", str(file_data["user"]))
                if "size" in data["files"][lfn]:
                    meta.add_attribute("size", str(file_data["size"]))
                if "ctime" in data["files"][lfn]:
                    meta.add_attribute("ctime", file_data["ctime"])
                if "sha256" in data["files"][lfn]:
                    meta.add_attribute("checksum.type", "sha256")
                    meta.add_attribute("checksum.value", file_data["sha256"])
                    if "checksum_timing" in data["files"][lfn]:
                        meta.add_attribute(
                            "checksum.timing", str(file_data["checksum_timing"])
                        )
                # what else?

                new_data["outputs"][lfn] = meta

        return new_data

    def parse_invocation_record(self, buffer=""):
        """
        Parses the YAML record in buffer returning an invocation record
        :param buffer:
        :return:
        """
        entry = {}

        # Check if we have an invocation record
        if self.is_invocation_record(buffer) is False:
            return entry

        try:
            entry = yaml.safe_load(buffer)[0]
        except Exception as e:
            logger.warning(
                "KICKSTART-PARSE-ERROR --> yaml error in %s : %s"
                % (self._kickstart_output_file, str(e))
            )

        # translate from the yaml dict structure to what we want using the keys-dict
        return self.map_yaml_to_ver2_format(entry)

    def parse_multipart_record(self, buffer=""):
        """
        Parses the YAML record in buffer returning a multipart record in the job.out file
        Sample buffer
            ---------------pegasus-multipart
            - integrity_verification_attempts:
              - lfn: "f.a"
                pfn: "f.a"
                sha256: 8e8ecb610e893781b6c0a38e443a257cb8c0aa548b04946930bea987e5e090d6
                success: True
            - integrity_summary:
                succeeded: 1
                failed: 0
                duration: 0.182

        :param buffer:
        :return: a list of yaml objects
        """
        entries = {}
        try:
            entries = yaml.safe_load(buffer)
        except Exception as e:
            logger.warning(
                "KICKSTART-PARSE-ERROR --> yaml error in multipart record %s : %s"
                % (self._kickstart_output_file, str(e))
            )

        for index, entry in enumerate(entries):
            entries[index]["multipart"] = True
        return entries


class XMLParser(Parser):
    """
    Represents the parser that parses the kickstart xml records
    """

    def __init__(self, filename):
        super().__init__(filename)
        self._parsing_job_element = False
        self._parsing_arguments = False
        self._parsing_main_job = False
        self._parsing_machine = False
        self._parsing_stdout = False
        self._parsing_stderr = False
        self._parsing_data = False
        self._parsing_cwd = False
        self._parsing_cpu = False
        self._parsing_final_statcall = False
        self._record_number = 0
        self._arguments = []
        self._stdout = ""
        self._stderr = ""
        self._cwd = ""
        self._lfn = ""  # filename parsed from statcall record

    def parse(self, keys_dict, tasks=True, clustered=True):
        """
        This function parses the previous XML based ks records,
        looking for
        the keys specified in the keys_dict variable. It returns a
        list of dictionaries containing the found keys. Look at the
        parse_stampede function for details about how to pass keys
        using the keys_dict structure. The function will return an
        empty list if no records are found or if an error happens.
        """
        my_reply = []

        # Place keys_dict in the _ks_elements
        self._ks_elements = keys_dict

        # Try to open the file
        if self.open() is False:
            return my_reply

        logger.debug(
            "Started reading records from kickstart file %s"
            % (self._kickstart_output_file)
        )

        self._record_number = 0
        # Read first record
        my_buffer = self.read_record()

        # Loop while we still have record to read
        while my_buffer is not None:
            if self.is_invocation_record(my_buffer) is True:
                # We have an invocation record, parse it!
                try:
                    my_record = self.parse_invocation_record(my_buffer)
                except Exception:
                    logger.warning(
                        "KICKSTART-PARSE-ERROR --> error parsing invocation record in file %s"
                        % (self._kickstart_output_file)
                    )
                    logger.warning(traceback.format_exc())
                    # Found error parsing this file, return empty reply
                    my_reply = []
                    # Finish the loop
                    break
                my_reply.append(my_record)
            elif self.is_clustered_record(my_buffer) is True:
                # Check if we want clustered records too
                if clustered:
                    # Clustered records are seqexec summary records for clustered jobs
                    # We have a clustered record, parse it!
                    my_reply.append(self.parse_clustered_record(my_buffer))
            elif self.is_task_record(my_buffer) is True:
                # Check if we want task records too
                if tasks:
                    # We have a clustered record, parse it!
                    my_reply.append(self.parse_task_record(my_buffer))
            else:
                # We have something else, this shouldn't happen!
                # Just skip it
                pass

            # Read next record
            my_buffer = self.read_record()

        # Lastly, close the file
        self.close()

        return my_reply

    def parse_invocation_record(self, buffer=""):
        """
        Parses the xml record in buffer, returning the desired keys.
        """
        # Initialize variables
        self._parsing_arguments = False
        self._parsing_main_job = False
        self._parsing_machine = False
        self._parsing_stdout = False
        self._parsing_stderr = False
        self._parsing_data = False
        self._parsing_cwd = False
        self._parsing_cpu = False
        self._parsing_signalled = False
        self._arguments = []
        self._stdout = ""
        self._stderr = ""
        self._cwd = ""
        self._keys = {}

        # Check if we have an invocation record
        if self.is_invocation_record(buffer) is False:
            return self._keys

        # Add invocation key to our response
        self._keys["invocation"] = True

        # Prepend XML header
        buffer = '<?xml version="1.0" encoding="ISO-8859-1"?>\n' + buffer

        # Create parser
        self._my_parser = expat.ParserCreate()
        self._my_parser.StartElementHandler = self.start_element
        self._my_parser.EndElementHandler = self.end_element
        self._my_parser.CharacterDataHandler = self.char_data

        # Parse everything!
        self._my_parser.Parse(buffer)

        # Add cwd, arguments, stdout, and stderr to keys
        if "cwd" in self._ks_elements:
            self._keys["cwd"] = self._cwd

        if "argument-vector" in self._ks_elements:
            self._keys["argument-vector"] = " ".join(self._arguments)

        if "stdout" in self._ks_elements:
            self._keys["stdout"] = self._stdout

        if "stderr" in self._ks_elements:
            self._keys["stderr"] = self._stderr

        return self._keys

    def read_record(self):
        """
        This function reads an invocation record from the kickstart
        output file. We also look for the struct at the end of a file
        containing multiple records. It returns a string containing
        the record, or None if it is not found.
        """
        buffer = ""

        # valid token that is parsed
        token = ""

        self._record_number += 1
        logger.debug(
            "Started reading record number %d from kickstart file %s"
            % (self._record_number, self._kickstart_output_file)
        )

        # First, we find the beginning <invocation xmlns....
        while True:
            line = self._fh.readline()
            if line == "":
                # End of file, record not found
                return None
            if line.find("<invocation") != -1:
                token = "<invocation"
                break
            if line.find("[cluster-task") != -1:
                token = "[cluster-task"
                break
            if line.find("[cluster-summary") != -1:
                token = "[cluster-summary"
                break
            if line.find("[seqexec-task") != -1:
                # deprecated token
                token = "[seqexec-task"
                break
            if line.find("[seqexec-summary") != -1:
                # deprecated token
                token = "[seqexec-summary"
                break

        # Found something!
        # if line.find("<invocation") >= 0:
        if token == "<invocation":
            # Found invocation record
            start = line.find("<invocation")
            buffer = line[start:]
            end = buffer.find("</invocation>")

            # Check if we have everything in a single line
            if end >= 0:
                end = end + len("</invocation>")
                return buffer[:end]
        # elif line.find("[seqexec-summary") >= 0:
        elif token == "[cluster-summary" or token == "[seqexec-summary":
            # Found line with cluster jobs summary
            start = line.find(token)
            buffer = line[start:]
            end = buffer.find("]")

            if end >= 0:
                end = end + len("]")
                return buffer[:end]

            # clustered record should be in a single line!
            logger.warning(
                "%s: %s line is malformed... ignoring it..."
                % (self._kickstart_output_file, token)
            )
            return ""
        # elif line.find("[seqexec-task") >= 0:
        elif token == "[cluster-task" or token == "[seqexec-task":
            # Found line with task information
            start = line.find(token)
            buffer = line[start:]
            end = buffer.find("]")

            if end >= 0:
                end = end + len("]")
                return buffer[:end]

            # task record should be in a single line!
            logger.warning(
                "%s: %s line is malformed... ignoring it..."
                % (self._kickstart_output_file, token)
            )
            return ""
        else:
            return ""

        # Ok, now continue reading the file until we get a full record
        buffer = [buffer]

        while True:
            line = self._fh.readline()
            if line == "":
                # End of file, record not found
                return None
            # buffer = buffer + line
            buffer.append(line)
            if line.find("</invocation>") >= 0:
                break

        # Now, we got it, let's make sure
        end = line.find("</invocation>")
        if end == -1:
            return ""

        # end = end + len("</invocation>")
        invocation = "".join(buffer)
        logger.debug(
            "Finished reading record number %d from kickstart file %s"
            % (self._record_number, self._kickstart_output_file)
        )
        return invocation
        # return buffer[:end]

    def start_element(self, name, attrs):
        """
        Function called by the parser every time a new element starts
        """
        # Keep track if we are parsing the main job element
        if name == "mainjob":
            self._parsing_main_job = True
        if name == "machine":
            self._parsing_machine = True
        # Keep track if we are inside one of the job elements
        if (
            name == "setup"
            or name == "prejob"
            or name == "mainjob"
            or name == "postjob"
            or name == "cleanup"
        ):
            self._parsing_job_element = True
        if name == "argument-vector" and name in self._ks_elements:
            # Start parsing arguments
            self._parsing_arguments = True
        elif name == "cwd" and name in self._ks_elements:
            # Start parsing cwd
            self._parsing_cwd = True
        elif name == "checksum" and name in self._ks_elements:
            # PM-1180 <checksum type="sha256" value="f2307670158c64c4407971f8fad67772724b0bad92bfb48f386b0814ba24e3af"/>
            self._keys[name] = {}
            for attr_name in self._ks_elements[name]:
                if attr_name in attrs:
                    self._keys[name][attr_name] = attrs[attr_name]
        elif name == "cpu" and name in self._ks_elements:
            # PM-1398 <cpu count="4" speed="2600" vendor="GenuineIntel">Intel(R) Xeon(R) CPU E5-2690 v4 @ 2.60GHz</cpu>
            self._parsing_cpu = True
            self._keys[name] = {}
            for attr_name in self._ks_elements[name]:
                if attr_name in attrs:
                    # keep consitency with 5.0 yaml based naming
                    self._keys[name]["cpu_" + attr_name] = attrs[attr_name]
        elif name == "data":
            # Start parsing data for stdout and stderr output
            self._parsing_data = True
        elif name == "file" and name in self._ks_elements:
            if self._parsing_main_job is True:
                # Special case for name inside the mainjob element (will change this later)
                for my_element in self._ks_elements[name]:
                    if my_element in attrs:
                        self._keys[my_element] = attrs[my_element]
        elif name == "ram" and name in self._ks_elements:
            if self._parsing_machine is True:
                # Special case for ram inside the machine element (will change this later)
                for my_element in self._ks_elements[name]:
                    if my_element in attrs:
                        self._keys[my_element] = attrs[my_element]
        elif name == "uname" and name in self._ks_elements:
            if self._parsing_machine is True:
                # Special case for uname inside the machine element (will change this later)
                for my_element in self._ks_elements[name]:
                    if my_element in attrs:
                        self._keys[my_element] = attrs[my_element]
        elif name == "signalled":
            # PM-1109 grab the attributes we are interested in
            self._keys[name] = {}  # a dictionary indexed by attributes
            self._parsing_signalled = True
            self._keys[name]["action"] = ""  # grabbed later in char data
            for attr in attrs:
                if attr in self._ks_elements[name]:
                    self._keys[name][attr] = attrs[attr]
        elif name == "statcall":
            if "id" in attrs:
                if attrs["id"] == "stdout" and "stdout" in self._ks_elements:
                    self._parsing_stdout = True
                elif attrs["id"] == "stderr" and "stderr" in self._ks_elements:
                    self._parsing_stderr = True
                elif attrs["id"] == "final":
                    self._parsing_final_statcall = True
                    self._lfn = attrs["lfn"]
        elif name == "statinfo":
            if self._parsing_final_statcall is True:
                statinfo = FileMetadata()
                for my_element in self._ks_elements[name]:
                    if my_element in attrs:
                        statinfo.add_attribute(my_element, attrs[my_element])
                if "outputs" not in self._keys:
                    self._keys["outputs"] = {}  # a dictionary indexed by lfn
                lfn = self._lfn
                statinfo.set_id(lfn)
                if lfn is None or not statinfo:
                    logger.warning(
                        "Malformed/Empty stat record for output lfn %s %s"
                        % (lfn, statinfo)
                    )
                self._keys["outputs"][lfn] = statinfo
        elif name == "usage" and name in self._ks_elements:
            if self._parsing_main_job:
                # Special case for handling utime and stime, which need to be added
                for my_element in self._ks_elements[name]:
                    try:
                        self._keys[my_element] = float(attrs[my_element])
                    except ValueError:
                        logger.warning(
                            "cannot convert element %s to float!" % (my_element)
                        )
        else:
            # For all other elements, check if we want them
            if name in self._ks_elements:
                for my_element in self._ks_elements[name]:
                    if my_element in attrs:
                        self._keys[my_element] = attrs[my_element]

    def end_element(self, name):
        """
        Function called by the parser whenever we reach the end of an element
        """
        # Stop parsing argement-vector and cwd if we reached the end of those elements
        if name == "argument-vector":
            self._parsing_arguments = False
        elif name == "cwd":
            self._parsing_cwd = False
        elif name == "cpu":
            self._parsing_cpu = False
        elif name == "mainjob":
            self._parsing_main_job = False
        elif name == "machine":
            self._parsing_machine = False
        elif name == "signalled":
            self._parsing_signalled = False
        elif name == "statcall":
            if self._parsing_stdout is True:
                self._parsing_stdout = False
            if self._parsing_stderr is True:
                self._parsing_stderr = False
            if self._parsing_final_statcall is True:
                self._parsing_final_statcall = False
            if "outputs" in self._keys:
                if self._lfn in self._keys["outputs"]:
                    # PM-1180 get the statinfo and update with checksum
                    statinfo = self._keys["outputs"][self._lfn]
                    if "checksum" in self._keys:
                        for key in self._keys["checksum"]:
                            statinfo.add_attribute(
                                "checksum." + key, self._keys["checksum"][key]
                            )
                        self._keys["checksum"] = {}

        elif name == "data":
            self._parsing_data = False
        # Now, see if we left one of the job elements
        if (
            name == "setup"
            or name == "prejob"
            or name == "mainjob"
            or name == "postjob"
            or name == "cleanup"
        ):
            self._parsing_job_element = False

    def char_data(self, data=""):
        """
        Function called by the parser whenever there's character data in an element
        """
        if self._parsing_cwd is True:
            self._cwd += data

        if self._parsing_cpu is True:
            if "cpu_model" not in self._keys["cpu"]:
                self._keys["cpu"]["cpu_model"] = ""
            self._keys["cpu"]["cpu_model"] += data

        elif self._parsing_arguments is True:
            self._arguments.append(data.strip())

        elif self._parsing_stdout is True and self._parsing_data is True:
            self._stdout += data

        elif self._parsing_stderr is True and self._parsing_data is True:
            self._stderr += data

        elif self._parsing_signalled is True:
            self._keys["signalled"]["action"] += data


if __name__ == "__main__":

    # Let's run a test!
    print("Testing kickstart output file parsing...")

    # log to the console
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    logger.addHandler(console)
    logger.debug("Logger has been configured")

    # Make sure we have an argument
    if len(sys.argv) < 2:
        print("For testing, please give a kickstart output filename!")
        sys.exit(1)

    # Create parser class
    p = Parser(sys.argv[1])

    # Parse file according to the Stampede schema
    output = p.parse_stampede()

    # Print output
    for record in output:
        pprint(record)



__author__ = "Rajiv Mayani"

import logging

from Pegasus.db import connection
from Pegasus.db.admin.admin_loader import DBAdminError
from Pegasus.db.errors import StampedeDBNotFoundError
from Pegasus.db.schema import *


# Main stats class.
class StampedeWorkflowStatistics:
    def __init__(self, connString=None, expand_workflow=True):
        self.log = logging.getLogger(
            "{}.{}".format(self.__module__, self.__class__.__name__)
        )
        try:
            self.session = connection.connect(connString)
        except (connection.ConnectionError, DBAdminError) as e:
            self.log.exception(e)
            raise StampedeDBNotFoundError

        self._expand = expand_workflow

        self._root_wf_id = []
        self._root_wf_uuid = []
        self.all_workflows = None
        self._job_filter_mode = None
        self._time_filter_mode = None
        self._host_filter = None
        self._xform_filter = {"include": None, "exclude": None}

        self._wfs = []

    def initialize(self, root_wf_uuid=None, root_wf_id=None):
        if root_wf_uuid is None and root_wf_id is None:
            self.log.error("Either root_wf_uuid or root_wf_id is required")
            raise ValueError("Either root_wf_uuid or root_wf_id is required")

        if root_wf_uuid == "*" or root_wf_id == "*":
            self.all_workflows = True

        q = self.session.query(Workflow.root_wf_id, Workflow.wf_id, Workflow.wf_uuid)

        if root_wf_uuid:
            if root_wf_uuid == "*":
                q = q.filter(Workflow.root_wf_id == Workflow.wf_id)
            else:
                q = q.filter(Workflow.wf_uuid.in_(root_wf_uuid))
        else:
            if root_wf_id == "*":
                q = q.filter(Workflow.root_wf_id == Workflow.wf_id)
            else:
                q = q.filter(Workflow.root_wf_id.in_(root_wf_id))

        result = q.all()

        for workflow in result:
            if workflow.root_wf_id != workflow.wf_id:
                self.log.error(
                    "Only root level workflows are supported %s", workflow.wf_uuid
                )
                raise ValueError(
                    "Only root level workflows are supported %s", workflow.wf_uuid
                )

            self._root_wf_id.append(workflow.wf_id)
            self._root_wf_uuid.append(workflow.wf_uuid)

        if (
            root_wf_uuid
            and root_wf_uuid != "*"
            and len(root_wf_uuid) != len(self._root_wf_uuid)
        ):
            self.log.error("Some workflows were not found")
            raise ValueError("Some workflows were not found")
        elif (
            root_wf_id
            and root_wf_id != "*"
            and len(root_wf_id) != len(self._root_wf_id)
        ):
            self.log.error("Some workflows were not found")
            raise ValueError("Some workflows were not found")

        # Initialize filters with default value
        self.set_job_filter()
        self.set_time_filter()
        self.set_host_filter()
        self.set_transformation_filter()

        return True

    def __filter_all(self, q, w=None):

        if w is None:
            w = Workflow

        if not self.all_workflows:
            q = q.filter(w.root_wf_id.in_(self._root_wf_id))

        return q

    def __filter_roots_only(self, q, w=None):

        if w is None:
            w = Workflow

        if self.all_workflows:
            q = q.filter(w.root_wf_id == w.wf_id)
        else:
            q = q.filter(w.wf_id.in_(self._root_wf_id))

        return q

    def get_workflow_ids(self):
        q = self.session.query(Workflow.root_wf_id, Workflow.wf_id, Workflow.wf_uuid)
        q = self.__filter_all(q)
        q = q.order_by(Workflow.root_wf_id)
        return q.all()

    def get_workflow_details(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Workflowdetails
        """
        q = self.session.query(
            Workflow.wf_id,
            Workflow.wf_uuid,
            Workflow.parent_wf_id,
            Workflow.root_wf_id,
            Workflow.dag_file_name,
            Workflow.submit_hostname,
            Workflow.submit_dir,
            Workflow.planner_arguments,
            Workflow.user,
            Workflow.grid_dn,
            Workflow.planner_version,
            Workflow.dax_label,
            Workflow.dax_version,
        )

        q = self.__filter_roots_only(q)

        return q.all()

    def get_total_tasks_status(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totaltask
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totaltasks
        """
        q = self.session.query(Task.task_id)

        if self._expand:
            q = self.__filter_all(q)
        else:
            q = self.__filter_roots_only(q)

        q = q.filter(Task.wf_id == Workflow.wf_id)
        q = q.filter(Task.job_id == Job.job_id)

        if self._get_job_filter(Task) is not None:
            q = q.filter(self._get_job_filter(Task))

        return q.count()

    def _base_task_statistics_query(self, success=True, pmc=False):
        w = orm.aliased(Workflow, name="w")
        j = orm.aliased(Job, name="j")
        ji = orm.aliased(JobInstance, name="ji")

        sq_1 = self.session.query(
            w.wf_id,
            j.job_id,
            ji.job_instance_id.label("jiid"),
            ji.job_submit_seq.label("jss"),
            func.max(ji.job_submit_seq).label("maxjss"),
        )
        if pmc:
            sq_1 = self.session.query(
                w.wf_id,
                j.job_id,
                ji.job_instance_id.label("jiid"),
                ji.job_submit_seq.label("jss"),
            )

        sq_1 = sq_1.join(j, w.wf_id == j.wf_id)
        sq_1 = sq_1.join(ji, j.job_id == ji.job_id)

        if self._expand:
            sq_1 = self.__filter_all(sq_1, w)
        else:
            sq_1 = self.__filter_roots_only(sq_1, w)

        if not pmc:
            sq_1 = sq_1.group_by(j.job_id)
        if self._get_job_filter(j) is not None:
            sq_1 = sq_1.filter(self._get_job_filter(j))
        sq_1 = sq_1.subquery("t")

        # PM-713 - Change to func.count(distinct(Invocation.abs_task_id)) from func.count(Invocation.exitcode)
        sq_2 = self.session.query(
            sq_1.c.wf_id, func.count(distinct(Invocation.abs_task_id)).label("count")
        )
        sq_2 = sq_2.select_from(
            orm.join(sq_1, Invocation, sq_1.c.jiid == Invocation.job_instance_id)
        )
        if not pmc:
            sq_2 = sq_2.filter(sq_1.c.jss == sq_1.c.maxjss)

        sq_2 = sq_2.filter(Invocation.abs_task_id != None)  # noqa: E711
        if success:
            sq_2 = sq_2.filter(Invocation.exitcode == 0)
        else:
            sq_2 = sq_2.filter(Invocation.exitcode != 0)
        sq_2 = sq_2.group_by(sq_1.c.wf_id)
        return sq_2

    def _task_statistics_query_sum(self, success=True, pmc=False):
        s = self._base_task_statistics_query(success, pmc).subquery("tt")
        q = self.session.query(func.sum(s.c.count).label("task_count"))
        return q.one()[0] or 0

    def get_total_succeeded_tasks_status(self, pmc=False):
        return self._task_statistics_query_sum(True, pmc)

    def get_total_failed_tasks_status(self):
        return self._task_statistics_query_sum(False, False)

    def get_total_tasks_retries(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totaltaskretries
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totaltaskretries
        """
        sq_1 = self.session.query(
            Workflow.wf_id.label("wid"), Invocation.abs_task_id.label("tid")
        )

        if self._expand:
            sq_1 = self.__filter_all(sq_1)
        else:
            sq_1 = self.__filter_roots_only(sq_1)

        sq_1 = sq_1.filter(Job.wf_id == Workflow.wf_id)
        sq_1 = sq_1.filter(Invocation.wf_id == Workflow.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstance.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())
        sq_1 = sq_1.filter(JobInstance.job_instance_id == Invocation.job_instance_id)
        sq_1 = sq_1.filter(Invocation.abs_task_id != None)  # noqa: E711

        i = 0
        f = {}
        for row in sq_1.all():
            i += 1
            if row not in f:
                f[row] = True

        return i - len(f.keys())

    def get_total_jobs_status(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totaljobs
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totaljobs
        """
        q = self.session.query(Job.job_id)

        if self._expand:
            q = self.__filter_all(q)
        else:
            q = self.__filter_roots_only(q)

        q = q.filter(Job.wf_id == Workflow.wf_id)
        if self._get_job_filter() is not None:
            q = q.filter(self._get_job_filter())

        return q.count()

    def get_total_succeeded_failed_jobs_status(self, classify_error=False, tag=None):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totalsucceeded_failed_jobs
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totalsucceededfailedjobs
        """
        JobInstanceSub = orm.aliased(JobInstance, name="JobInstanceSub")
        sq_1 = self.session.query(
            func.max(JobInstanceSub.job_submit_seq).label("jss"),
            JobInstanceSub.job_id.label("jobid"),
        )

        if self._expand:
            sq_1 = self.__filter_all(sq_1)
        else:
            sq_1 = self.__filter_roots_only(sq_1)

        sq_1 = sq_1.filter(Workflow.wf_id == Job.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstanceSub.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())
        sq_1 = sq_1.group_by(JobInstanceSub.job_id).subquery()

        q = self.session.query(
            func.sum(case([(JobInstance.exitcode == 0, 1)], else_=0)).label(
                "succeeded"
            ),
            func.sum(case([(JobInstance.exitcode != 0, 1)], else_=0)).label("failed"),
        )
        q = q.filter(JobInstance.job_id == sq_1.c.jobid)
        q = q.filter(JobInstance.job_submit_seq == sq_1.c.jss)

        if classify_error:
            if tag is None:
                self.log.error("for error classification you need to specify tag")
                return None

            q = q.filter(JobInstance.job_instance_id == Tag.job_instance_id)
            q = q.filter(Tag.name == tag)
            q = q.filter(Tag.count > 0)

        return q.one()

    def get_total_jobs_retries(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-TotalJobRetries
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-TotalJobRetries
        """
        self._dax_or_dag_cond()

        sq_1 = self.session.query(func.count(Job.job_id))

        if self._expand:
            sq_1 = self.__filter_all(sq_1)
        else:
            sq_1 = self.__filter_roots_only(sq_1)

        sq_1 = sq_1.filter(Job.wf_id == Workflow.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstance.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())

        sq_1 = sq_1.subquery()

        sq_2 = self.session.query(func.count(distinct(JobInstance.job_id)))

        if self._expand:
            sq_2 = self.__filter_all(sq_2)
        else:
            sq_2 = self.__filter_roots_only(sq_2)

        sq_2 = sq_2.filter(Job.wf_id == Workflow.wf_id)
        sq_2 = sq_2.filter(Job.job_id == JobInstance.job_id)
        if self._get_job_filter() is not None:
            sq_2 = sq_2.filter(self._get_job_filter())

        sq_2 = sq_2.subquery()

        q = self.session.query(
            (sq_1.as_scalar() - sq_2.as_scalar()).label("total_job_retries")
        )

        return q.all()[0].total_job_retries

    def get_workflow_states(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Workflowwalltime
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Workflowwalltime
        """
        q = self.session.query(
            Workflowstate.wf_id,
            Workflowstate.state,
            Workflowstate.timestamp,
            Workflowstate.restart_count,
            Workflowstate.status,
        )

        if self.all_workflows:
            q = q.filter(Workflowstate.wf_id == Workflow.wf_id)
            q = q.filter(Workflow.root_wf_id == Workflow.wf_id)
        else:
            q = q.filter(Workflowstate.wf_id.in_(self._root_wf_id))

        q = q.order_by(Workflowstate.wf_id, Workflowstate.restart_count)

        return q.all()

    def get_workflow_cum_job_wall_time(self):
        """
        select sum(remote_duration * multiplier_factor) FROM
        invocation as invoc, job_instance as ji WHERE
        invoc.task_submit_seq >= 0 and
        invoc.job_instance_id = ji.job_instance_id and
        invoc.wf_id in (1,2,3) and
        invoc.transformation <> 'condor::dagman'

        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Workflowcumulativejobwalltime
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Workflowcumulativejobwalltime
        """
        q = self.session.query(
            cast(
                func.sum(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ),
            cast(
                func.sum(
                    case(
                        [
                            (
                                Invocation.exitcode == 0,
                                Invocation.remote_duration
                                * JobInstance.multiplier_factor,
                            )
                        ],
                        else_=0,
                    )
                ).label("goodput"),
                Float,
            ),
            cast(
                func.sum(
                    case(
                        [
                            (
                                Invocation.exitcode > 0,
                                Invocation.remote_duration
                                * JobInstance.multiplier_factor,
                            )
                        ],
                        else_=0,
                    )
                ).label("badput"),
                Float,
            ),
        )

        q = q.filter(Invocation.task_submit_seq >= 0)
        q = q.filter(Invocation.job_instance_id == JobInstance.job_instance_id)

        if self._expand:
            q = self.__filter_all(q)
            q = q.filter(Invocation.wf_id == Workflow.wf_id)
        else:
            if self.all_workflows:
                q = self.__filter_roots_only(q)
                q = q.filter(Invocation.wf_id == Workflow.wf_id)
            else:
                q = q.filter(Invocation.wf_id.in_(self._root_wf_id))

        q = q.filter(Invocation.transformation != "condor::dagman")
        return q.first()

    def get_summary_integrity_metrics(self):
        """

        :param type:    whether integrity type is check | compute
        :param file_type: file type input or output
        :return:
        """
        q = self.session.query(
            IntegrityMetrics.type,
            func.sum(IntegrityMetrics.duration).label("duration"),
            func.sum(IntegrityMetrics.count).label("count"),
        )

        q = q.group_by(IntegrityMetrics.type)

        if self._expand:
            q = self.__filter_all(q)
            q = q.filter(IntegrityMetrics.wf_id == Workflow.wf_id)
        else:
            if self.all_workflows:
                q = self.__filter_roots_only(q)
                q = q.filter(IntegrityMetrics.wf_id == Workflow.wf_id)
            else:
                q = q.filter(IntegrityMetrics.wf_id.in_(self._wfs))

        # at most two records grouped by type compute | check
        return q.all()

    def get_tag_metrics(self, name):
        """

        :param name:    what type of tag to aggregate on
        :return:
        """
        q = self.session.query(Tag.name, func.sum(Tag.count).label("count"))

        q = q.group_by(Tag.name)
        q = q.filter(Tag.name == name)

        if self._expand:
            q = self.__filter_all(q)
            q = q.filter(Tag.wf_id == Workflow.wf_id)
        else:
            if self.all_workflows:
                q = self.__filter_roots_only(q)
                q = q.filter(Tag.wf_id == Workflow.wf_id)
            else:
                q = q.filter(Tag.wf_id.in_(self._wfs))

        return q.all()

    def get_integrity_metrics(self):
        """

        :param type:    whether integrity type is check | compute
        :param file_type: file type input or output
        :return:
        """
        q = self.session.query(
            IntegrityMetrics.type,
            IntegrityMetrics.file_type,
            func.sum(IntegrityMetrics.duration).label("duration"),
            func.sum(IntegrityMetrics.count).label("count"),
        )

        q = q.group_by(IntegrityMetrics.type)
        q = q.group_by(IntegrityMetrics.file_type)

        if self._expand:
            q = self.__filter_all(q)
            q = q.filter(IntegrityMetrics.wf_id == Workflow.wf_id)
        else:
            if self.all_workflows:
                q = self.__filter_roots_only(q)
                q = q.filter(IntegrityMetrics.wf_id == Workflow.wf_id)
            else:
                q = q.filter(IntegrityMetrics.wf_id.in_(self._wfs))

        return q.all()

    def get_submit_side_job_wall_time(self):
        """
        select sum(local_duration * multiplier_factor) FROM
        job_instance as jb_inst, job as jb WHERE
        jb_inst.job_id  = jb.job_id and
        jb.wf_id in (1,2,3) and
        ((not (jb.type_desc ='dax' or jb.type_desc ='dag'))
          or
         ((jb.type_desc ='dax' or jb.type_desc ='dag') and jb_inst.subwf_id is NULL)
        )

        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Cumulativejobwalltimeasseenfromsubmitside
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Cumulativejobwalltimeasseenfromsubmitside
        """
        q = self.session.query(
            cast(
                func.sum(JobInstance.local_duration * JobInstance.multiplier_factor),
                Float,
            ).label("wall_time"),
            cast(
                func.sum(
                    case(
                        [
                            (
                                JobInstance.exitcode == 0,
                                JobInstance.local_duration
                                * JobInstance.multiplier_factor,
                            )
                        ],
                        else_=0,
                    )
                ).label("goodput"),
                Float,
            ),
            cast(
                func.sum(
                    case(
                        [
                            (
                                JobInstance.exitcode > 0,
                                JobInstance.local_duration
                                * JobInstance.multiplier_factor,
                            )
                        ],
                        else_=0,
                    )
                ).label("badput"),
                Float,
            ),
        )

        q = q.filter(JobInstance.job_id == Job.job_id)

        if self._expand:
            q = self.__filter_all(q)
            q = q.filter(Job.wf_id == Workflow.wf_id)
        else:
            if self.all_workflows:
                q = self.__filter_roots_only(q)
                q = q.filter(Job.wf_id == Workflow.wf_id)
            else:
                q = q.filter(Job.wf_id.in_(self._root_wf_id))

        if self._expand:
            d_or_d = self._dax_or_dag_cond()
            q = q.filter(
                or_(not_(d_or_d), and_(d_or_d, JobInstance.subwf_id == None))
            )  # noqa: E711

        return q.first()

    def get_workflow_retries(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Workflowretries
        """
        sq_1 = self.session.query(func.max(Workflowstate.restart_count).label("retry"))

        if self._expand:
            sq_1 = self.__filter_all(sq_1)
        else:
            sq_1 = self.__filter_roots_only(sq_1)

        sq_1 = sq_1.filter(Workflowstate.wf_id == Workflow.wf_id)
        sq_1 = sq_1.group_by(Workflowstate.wf_id)
        sq_1 = sq_1.subquery()

        q = self.session.query(func.sum(sq_1.c.retry).label("total_retry"))

        return q.one().total_retry

    def get_transformation_statistics(self):
        """
        SELECT transformation,
               count(invocation_id) as count,
               min(remote_duration * multiplier_factor) as min,
               count(CASE WHEN (invoc.exitcode = 0 and invoc.exitcode is NOT NULL) THEN invoc.exitcode END) AS success,
               count(CASE WHEN (invoc.exitcode != 0 and invoc.exitcode is NOT NULL) THEN invoc.exitcode END) AS failure,
               max(remote_duration * multiplier_factor) as max,
               avg(remote_duration * multiplier_factor) as avg,
               sum(remote_duration * multiplier_factor) as sum FROM
        invocation as invoc, job_instance as ji WHERE
        invoc.job_instance_id = ji.job_instance_id and
        invoc.wf_id IN (1,2,3) GROUP BY transformation
        """
        q = self.session.query(
            Invocation.transformation,
            func.count(Invocation.invocation_id).label("count"),
            cast(
                func.min(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ).label("min"),
            func.count(case([(Invocation.exitcode == 0, Invocation.exitcode)])).label(
                "success"
            ),
            func.count(case([(Invocation.exitcode != 0, Invocation.exitcode)])).label(
                "failure"
            ),
            cast(
                func.max(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ).label("max"),
            cast(
                func.avg(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ).label("avg"),
            cast(
                func.sum(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ).label("sum"),
        )
        q = q.filter(Workflow.wf_id == Invocation.wf_id)
        q = q.filter(Invocation.job_instance_id == JobInstance.job_instance_id)
        q = self.__filter_all(q)
        q = q.group_by(Invocation.transformation)

        return q.all()

    def get_invocation_by_time(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
        """
        q = self.session.query(
            (cast(Invocation.start_time / self._get_date_divisors(), Integer)).label(
                "date_format"
            ),
            func.count(Invocation.invocation_id).label("count"),
            cast(func.sum(Invocation.remote_duration), Float).label("total_runtime"),
        )
        q = self.__filter_all(q)
        q = q.filter(Invocation.wf_id == Workflow.wf_id)
        if self._get_xform_filter() is not None:
            q = q.filter(self._get_xform_filter())
        q = q.group_by("date_format").order_by("date_format")

        return q.all()

    def get_invocation_by_time_per_host(self, host=None):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
        """
        q = self.session.query(
            (cast(Invocation.start_time / self._get_date_divisors(), Integer)).label(
                "date_format"
            ),
            Host.hostname.label("host_name"),
            func.count(Invocation.invocation_id).label("count"),
            cast(func.sum(Invocation.remote_duration), Float).label("total_runtime"),
        )
        q = self.__filter_all(q)
        q = q.filter(Invocation.wf_id == Workflow.wf_id)
        q = q.filter(JobInstance.job_instance_id == Invocation.job_instance_id)
        q = q.filter(JobInstance.host_id == Host.host_id)
        if self._get_host_filter() is not None:
            q = q.filter(self._get_host_filter())
        if self._get_xform_filter() is not None:
            q = q.filter(self._get_xform_filter())
        q = q.group_by("date_format", "host_name").order_by("date_format")

        return q.all()

    def get_jobs_run_by_time_per_host(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
        """
        q = self.session.query(
            (cast(Jobstate.timestamp / self._get_date_divisors(), Integer)).label(
                "date_format"
            ),
            Host.hostname.label("host_name"),
            func.count(JobInstance.job_instance_id).label("count"),
            cast(func.sum(JobInstance.local_duration), Float).label("total_runtime"),
        )
        q = self.__filter_all(q)
        q = q.filter(Workflow.wf_id == Job.wf_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(JobInstance.job_instance_id == Jobstate.job_instance_id)
        q = q.filter(JobInstance.host_id == Host.host_id)
        q = q.filter(Jobstate.state == "EXECUTE")

        if self._get_host_filter() is not None:
            q = q.filter(self._get_host_filter())
        if self._get_job_filter() is not None:
            q = q.filter(self._get_job_filter())

        q = q.group_by("date_format", "host_name").order_by("date_format")

        return q.all()

    def get_jobs_run_by_time(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
        """
        q = self.session.query(
            (cast(Jobstate.timestamp / self._get_date_divisors(), Integer)).label(
                "date_format"
            ),
            func.count(JobInstance.job_instance_id).label("count"),
            cast(func.sum(JobInstance.local_duration), Float).label("total_runtime"),
        )
        q = q.filter(Workflow.wf_id == Job.wf_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(JobInstance.job_instance_id == Jobstate.job_instance_id)

        q = self.__filter_all(q)
        q = q.filter(Jobstate.state == "EXECUTE")
        q = q.filter(JobInstance.local_duration != None)  # noqa: E711

        if self._get_job_filter() is not None:
            q = q.filter(self._get_job_filter())
        q = q.group_by("date_format").order_by("date_format")

        return q.all()

    def close(self):
        self.log.debug("close")
        self.session.close()

    def set_job_filter(self, filter="all"):
        modes = [
            "all",
            "nonsub",
            "subwf",
            "dax",
            "dag",
            "compute",
            "stage-in-tx",
            "stage-out-tx",
            "registration",
            "inter-site-tx",
            "create-dir",
            "staged-compute",
            "cleanup",
            "chmod",
        ]
        try:
            modes.index(filter)
            self._job_filter_mode = filter
            self.log.debug("Setting filter to: %s", filter)
        except Exception:
            self._job_filter_mode = "all"
            self.log.error("Unknown job filter %s - setting to all", filter)

    def set_time_filter(self, filter="month"):
        modes = ["month", "week", "day", "hour"]
        try:
            modes.index(filter)
            self._time_filter_mode = filter
            self.log.debug("Setting filter to: %s", filter)
        except Exception:
            self._time_filter_mode = "month"
            self.log.error("Unknown time filter %s - setting to month", filter)

    def set_host_filter(self, host=None):
        """
        The host argument can either be a string/single hostname or
        it can be a list/array of hostnames.
        """
        self._host_filter = host

    def set_transformation_filter(self, include=None, exclude=None):
        """
        Either of these args can either be a single string/xform type or
        it can be a list/array of xform types.

        Both arguments can not be set at the same time.  If they are,
        the program will log an error and not do any filtering.
        """
        self._xform_filter["include"] = include
        self._xform_filter["exclude"] = exclude

    def get_schema_version(self):
        return self.s_check.check_version()

    #
    # Status of initially planned wf components.
    #

    #
    # The following block of queries are documented here:
    # https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary
    # and
    # https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file
    #

    def _dax_or_dag_cond(self, JobO=Job):
        return or_(JobO.type_desc == "dax", JobO.type_desc == "dag")

    def _get_job_filter(self, JobO=Job):
        filters = {
            "all": None,
            "nonsub": not_(self._dax_or_dag_cond(JobO)),
            "subwf": self._dax_or_dag_cond(JobO),
            "dax": JobO.type_desc == "dax",
            "dag": JobO.type_desc == "dag",
            "compute": JobO.type_desc == "compute",
            "stage-in-tx": JobO.type_desc == "stage-in-tx",
            "stage-out-tx": JobO.type_desc == "stage-out-tx",
            "registration": JobO.type_desc == "registration",
            "inter-site-tx": JobO.type_desc == "inter-site-tx",
            "create-dir": JobO.type_desc == "create-dir",
            "staged-compute": JobO.type_desc == "staged-compute",
            "cleanup": JobO.type_desc == "cleanup",
            "chmod": JobO.type_desc == "chmod",
        }
        return filters[self._job_filter_mode]

    #
    # Runtime queries
    # https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
    #

    def _get_date_divisors(self):
        vals = {"month": 2629743, "week": 604800, "day": 86400, "hour": 3600}
        return vals[self._time_filter_mode]

    def _get_host_filter(self):
        if self._host_filter is None:
            return None
        elif isinstance(self._host_filter, str):
            return Host.hostname == self._host_filter
        elif isinstance(self._host_filter, type([])):
            return Host.hostname.in_(self._host_filter)
        else:
            return None

    def _get_xform_filter(self):
        if (
            self._xform_filter["include"] is not None
            and self._xform_filter["exclude"] is not None
        ):
            self.log.error(
                "Can't set both transform include and exclude - reset s.set_transformation_filter()"
            )
            return None
        elif (
            self._xform_filter["include"] is None
            and self._xform_filter["exclude"] is None
        ):
            return None
        elif self._xform_filter["include"] is not None:
            if isinstance(self._xform_filter["include"], str):
                return Invocation.transformation == self._xform_filter["include"]
            elif isinstance(self._xform_filter["include"], type([])):
                return Invocation.transformation.in_(self._xform_filter["include"])
            else:
                return None
        elif self._xform_filter["exclude"] is not None:
            if isinstance(self._xform_filter["exclude"], str):
                return Invocation.transformation != self._xform_filter["exclude"]
            elif isinstance(self._xform_filter["exclude"], type([])):
                return not_(
                    Invocation.transformation.in_(self._xform_filter["exclude"])
                )
            else:
                return None
        else:
            return None



import logging
import os
import uuid
from datetime import datetime

from flask import Response, abort, current_app, g, make_response, request, url_for

from Pegasus import user
from Pegasus.service.auth import (  # noqa: F401
    BaseAuthentication,
    NoAuthentication,
    PAMAuthentication,
)

log = logging.getLogger(__name__)


def _get_request_uuid():
    """
    Generate UUID for a request.

    :return: first section of uuid generated using uuid.uuid4
    """
    uid = "%s" % uuid.uuid4()
    return uid.split("-")[0]


def _basic_auth_response():
    return Response(
        "Basic Auth Required",
        401,
        {"WWW-Authenticate": 'Basic realm="Pegasus Service"'},
    )


def _is_user_an_admin(username):
    """
    Check if user ia a valid admin user.
    """
    admin_users = current_app.config["ADMIN_USERS"]

    if isinstance(admin_users, str):
        admin_users = admin_users.strip()

    if admin_users is None or admin_users is False or admin_users == "":
        return False
    elif admin_users == "*":
        return True
    elif hasattr(admin_users, "__iter__"):
        return username in admin_users
    else:
        log.error("Invalid configuration: ADMIN_USERS is invalid.")
        abort(500)


def before_request():
    request.uid = _get_request_uuid()
    request.start_time = datetime.utcnow()


def is_xhr():
    request.is_xhr = (
        request.headers.get("X-Requested-With", "False") == "XMLHttpRequest"
    )


def authenticate():
    # Static files do not need to be authenticated.
    if (request.script_root + request.path).startswith(url_for("static", filename="")):
        return

    #
    # Authentication
    #

    cred = request.authorization
    username = cred.username if cred else None
    password = cred.password if cred else None

    authclass = current_app.config["AUTHENTICATION"]
    if authclass not in globals():
        log.error("Unknown authentication method: %s", authclass)
        return make_response("Invalid server configuration", 500)

    Authentication = globals()[authclass]
    auth = Authentication(username, password)
    if not auth.authenticate():
        log.error("Invalid login: %s", username)
        return _basic_auth_response()

    try:
        g.user = auth.get_user()
    except user.NoSuchUser:
        log.error("No such user: %s" % username)
        return _basic_auth_response()

    log.info("Authenticated user %s", g.user.username)

    # If a username is not specified in the requested URI, then set username to
    # the logged in user?
    if "username" not in g:
        g.username = g.user.username


def authorization():
    # Static files do not need to be authorized.
    if (request.script_root + request.path).startswith(url_for("static", filename="")):
        return

    # Root user is off limits.
    if g.username == "root":
        log.error("Accessing root user info. is not allowed")
        # If the user has logged in as root, then ask user to login as a regular user.
        # If the non-root logged in user is attempting to access root user's data,
        # then return 403 FORBIDDEN
        if g.user.username == "root":
            return _basic_auth_response()
        else:
            abort(403)

    user_info = g.user

    if g.username != g.user.username:
        # Is user (g.user.username) allowed to view user (g.username) runs?
        if not _is_user_an_admin(g.user.username):
            log.error(
                "User {} is accessing user {}'s runs".format(
                    g.user.username, g.username
                )
            )
            abort(403)

        # Is user a valid system user?
        try:
            user_info = user.get_user_by_username(g.username)
        except user.NoSuchUser:
            log.error("User %s is not a valid user" % g.username)
            abort(400)

    if current_app.config["PROCESS_SWITCHING"]:
        # If required, set uid and gid of handler process
        if os.getuid() != user_info.uid:
            if os.getuid() != 0:
                log.error(
                    "Pegasus service must run as root to enable process switching"
                )
                return make_response(
                    "Pegasus service must run as root to enable process switching", 500
                )

        os.setgid(user_info.gid)
        os.setuid(user_info.uid)

    # Does the user have a Pegasus home directory?
    user_pegasus_dir = user_info.get_pegasus_dir()

    if not os.path.isdir(user_pegasus_dir):
        log.info("User's pegasus directory does not exist. Creating one...")
        try:
            os.makedirs(user_pegasus_dir, mode=0o744)
        except OSError:
            log.info("Invalid Permissions: Could not create user's pegasus directory.")
            return make_response("Could not find user's Pegasus directory", 404)

    # Set master DB URL for the dashboard
    # For testing master_db_url would be pre-populated, so let's not overwrite it here.
    if "master_db_url" not in g:
        g.master_db_url = user_info.get_master_db_url()


def teardown_request(response):
    end_time = datetime.utcnow()
    start_time = getattr(request, "start_time", end_time)
    delta = end_time - start_time
    uid = getattr(request, "uid", "no-uid")

    log.debug("Request ID: {} took {} seconds".format(uid, delta.total_seconds()))

    return response


def add_username(endpoint, values):
    """
    If the endpoint expects a variable username, then set it's value to g.username.
    This is done so as not to provide g.username as a parameter to every call to url_for.
    """

    #
    # Route does not expects a value for username
    #

    if not current_app.url_map.is_endpoint_expecting(endpoint, "username"):
        return

    #
    # Route expects a value for username
    #

    # Value for username has already been provided
    if "username" in values or ("username" in g and not g.username):
        return

    values["username"] = g.username


def pull_username(endpoint, values):
    """
    If the requested endpoint contains a value for username variable then extract it and set it in g.username.
    """
    if values and "username" in values:
        g.username = values["username"]


def register_lifecycle_handlers(app):
    # For each request create a unique id and record the start time
    app.before_request(before_request)

    # Is Ajax request?
    app.before_request(is_xhr)

    # Authentication
    app.before_request(authenticate)

    # Authorization
    app.before_request(authorization)

    # After each request log the request duration
    app.teardown_request(teardown_request)

    app.url_defaults(add_username)
    app.url_value_preprocessor(pull_username)



#!/usr/bin/env python
#
#  Copyright 2018-2020 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

__author__ = "Monte Goode"
__author__ = "Karan Vahi"
__author__ = "Rafael Ferreira da Silva"
__author__ = "Rajiv Mayani"

import logging
import time
import warnings

from sqlalchemy.dialects import mysql, postgresql, sqlite
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import foreign, mapper, relation
from sqlalchemy.schema import Column, ForeignKey, Index, MetaData, UniqueConstraint
from sqlalchemy.sql.expression import and_
from sqlalchemy.types import (
    BigInteger,
    Boolean,
    DateTime,
    Enum,
    Integer,
    Numeric,
    String,
    Text,
)

from Pegasus.db.ensembles import Ensemble as _Ensemble
from Pegasus.db.ensembles import EnsembleStates
from Pegasus.db.ensembles import EnsembleWorkflow as _EnsembleWorkflow
from Pegasus.db.ensembles import EnsembleWorkflowStates, TriggerStates, TriggerType

__all__ = (
    "DBVersion",
    "Workflow",
    "Workflowstate",
    "WorkflowMeta",
    "Host",
    "Job",
    "JobEdge",
    "JobInstance",
    "Jobstate",
    "Tag",
    "Task",
    "TaskEdge",
    "TaskMeta",
    "Invocation",
    "WorkflowFiles",
    "IntegrityMetrics",
    "MasterWorkflow",
    "MasterWorkflowstate",
    "Ensemble",
    "EnsembleWorkflow",
    "Trigger",
    "RCLFN",
    "RCPFN",
    "RCMeta",
)

log = logging.getLogger(__name__)

# for SQLite
warnings.filterwarnings("ignore", r".*does \*not\* support Decimal*.")

# These are keywords that all tables should have
table_keywords = {"mysql_charset": "utf8mb4", "mysql_engine": "InnoDB"}

KeyInteger = BigInteger()
KeyInteger = KeyInteger.with_variant(postgresql.BIGINT(), "postgresql")
KeyInteger = KeyInteger.with_variant(mysql.BIGINT(), "mysql")
KeyInteger = KeyInteger.with_variant(sqlite.INTEGER(), "sqlite")

TimestampType = Numeric(precision=16, scale=6)
DurationType = Numeric(precision=10, scale=3)


# --------------------------------------------------------------------


class SABase:
    """
    Base class for all the DB mapper objects.
    """

    def _commit(self, session, batch, merge=False):
        if merge:
            session.merge(self)
        else:
            session.add(self)
        if batch:
            return
        session.flush()
        session.commit()

    def commit_to_db(self, session, batch=False):
        """
        Commit the DB object/row to the database.

        @type   session: sqlalchemy.orm.scoping.ScopedSession object
        @param  session: SQLAlch session to commit row to.
        """
        self._commit(session, batch)

    def merge_to_db(self, session, batch=False):
        """
        Merge the DB object/row with an existing row in the database.

        @type   session: sqlalchemy.orm.scoping.ScopedSession object
        @param  session: SQLAlch session to merge row with.

        Using this method pre-supposes that the developer has already
        assigned any primary key information to the object before
        calling.
        """
        self._commit(session, batch, merge=True)

    def __repr__(self):
        retval = "%s:\n" % self.__class__
        for k, v in self.__dict__.items():
            if k == "_sa_instance_state":
                continue
            retval += "  * {} : {}\n".format(k, v)
        return retval


metadata = MetaData()
Base = declarative_base(cls=SABase, metadata=metadata)


# ----------------------------------------------------------------
# Method to verify if tables exist or are according to the schema
# ----------------------------------------------------------------


def get_missing_tables(db):
    tables = (
        DBVersion,
        # WORKFLOW
        Workflow,
        Workflowstate,
        WorkflowMeta,
        WorkflowFiles,
        Host,
        Job,
        JobEdge,
        JobInstance,
        Jobstate,
        Tag,
        Task,
        TaskEdge,
        TaskMeta,
        Invocation,
        IntegrityMetrics,
        # MASTER
        MasterWorkflow,
        MasterWorkflowstate,
        Ensemble,
        EnsembleWorkflow,
        # JDBCRC
        RCLFN,
        RCPFN,
        RCMeta,
    )
    missing_tables = []
    for table in tables:
        if not check_table_exists(db, table):
            missing_tables.append(table.__tablename__)

    return missing_tables


def check_table_exists(engine, table):
    try:
        engine.execute(table.__table__.select().limit(1))
        return True

    except OperationalError as e:
        if (
            "no such table" in str(e).lower()
            or "unknown" in str(e).lower()
            or "no such column" in str(e).lower()
        ):
            return False
        raise
    except ProgrammingError as e:
        if "doesn't exist" in str(e).lower():
            return False
        raise


# ---------------------------------------------
# DB ADMIN
# ---------------------------------------------


class DBVersion(Base):
    """."""

    __tablename__ = "dbversion"
    __table_args__ = (
        {
            "mysql_charset": "utf8mb4",
            "mysql_engine": "InnoDB",
            "sqlite_autoincrement": True,
        },
    )

    id = Column("id", KeyInteger, primary_key=True, autoincrement=True)
    version_number = Column("version_number", Integer, default=5)
    version = Column("version", String(50), nullable=False)
    version_timestamp = Column("version_timestamp", Integer, nullable=False)


# ---------------------------------------------
# STAMPEDE
# ---------------------------------------------


class Workflow(Base):
    """."""

    __tablename__ = "workflow"

    # ==> Information comes from braindump.txt file
    wf_id = Column("wf_id", KeyInteger, primary_key=True)
    wf_uuid = Column("wf_uuid", String(255), nullable=False)
    dag_file_name = Column("dag_file_name", String(255))
    timestamp = Column("timestamp", TimestampType)
    submit_hostname = Column("submit_hostname", String(255))
    submit_dir = Column("submit_dir", Text)
    planner_arguments = Column("planner_arguments", Text)
    user = Column("user", String(255))
    grid_dn = Column("grid_dn", String(255))
    planner_version = Column("planner_version", String(255))
    dax_label = Column("dax_label", String(255))
    dax_version = Column("dax_version", String(255))
    dax_file = Column("dax_file", String(255))
    db_url = Column("db_url", Text)
    parent_wf_id = Column(
        "parent_wf_id", KeyInteger, ForeignKey(wf_id, ondelete="CASCADE"),
    )
    # not marked as FK to not screw up the cascade.
    root_wf_id = Column("root_wf_id", KeyInteger)

    # Relationships
    root_wf = relation(
        lambda: Workflow,
        cascade="all, delete-orphan",
        single_parent=True,
        remote_side=(wf_id,),
    )
    parent_wf = relation(
        lambda: Workflow,
        cascade="all, delete-orphan",
        single_parent=True,
        remote_side=(wf_id,),
    )
    states = relation(
        lambda: Workflowstate,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by=lambda: Workflowstate.timestamp,
    )
    jobs = relation(
        lambda: Job,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    job_edges = relation(
        lambda: JobEdge,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    tasks = relation(
        lambda: Task,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    task_edges = relation(
        lambda: TaskEdge,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    invocations = relation(
        lambda: Invocation,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    hosts = relation(
        lambda: Host,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    workflow_files = relation(
        lambda: WorkflowFiles,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    files = relation(
        lambda: RCLFN,
        secondary=lambda: WorkflowFiles.__table__,
        primaryjoin=lambda: Workflow.wf_id == WorkflowFiles.wf_id,
        secondaryjoin=lambda: WorkflowFiles.lfn_id == RCLFN.lfn_id,
    )
    integrity_metrics = relation(
        lambda: IntegrityMetrics,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    tags = relation(
        lambda: Tag,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    meta = relation(
        lambda: WorkflowMeta,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )


Workflow.__table_args__ = (
    UniqueConstraint(Workflow.wf_uuid, name="UNIQUE_WF_UUID"),
    table_keywords,
)


class Workflowstate(Base):
    """."""

    __tablename__ = "workflowstate"
    __table_args__ = (table_keywords,)

    # All three columns are marked as primary key to produce the desired
    # effect - ie: it is the combo of the three columns that make a row
    # unique.
    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        primary_key=True,
    )
    state = Column(
        "state",
        Enum("WORKFLOW_STARTED", "WORKFLOW_TERMINATED", name="workflow_state"),
        primary_key=True,
    )
    timestamp = Column(
        "timestamp", TimestampType, primary_key=True, default=time.time(),
    )
    restart_count = Column("restart_count", Integer, nullable=False)
    status = Column("status", Integer)
    reason = Column("reason", Text)


Index("workflowstate_timestamp_COL", Workflowstate.timestamp)


class WorkflowMeta(Base):
    """."""

    __tablename__ = "workflow_meta"
    __table_args__ = (table_keywords,)

    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        primary_key=True,
    )
    key = Column("key", String(255), primary_key=True)
    value = Column("value", String(255), nullable=False)


# Host definition
# ==> Information from kickstart output file
#
# site_name = <resource, from invocation element>
# hostname = <hostname, from invocation element>
# ip_address = <hostaddr, from invocation element>
# uname = <combined (system, release, machine) from machine element>
# total_ram = <ram_total from machine element>


class Host(Base):
    """."""

    __tablename__ = "host"

    host_id = Column("host_id", KeyInteger, primary_key=True)
    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        nullable=False,
    )
    site = Column("site", String(255), nullable=False)
    hostname = Column("hostname", String(255), nullable=False)
    ip = Column("ip", String(255), nullable=False)
    uname = Column("uname", String(255))
    total_memory = Column("total_memory", Integer)


Host.__table_args__ = (
    UniqueConstraint(Host.wf_id, Host.site, Host.hostname, Host.ip, name="UNIQUE_HOST"),
    table_keywords,
)


class Job(Base):
    """."""

    __tablename__ = "job"

    job_id = Column("job_id", KeyInteger, primary_key=True)
    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        nullable=False,
    )
    exec_job_id = Column("exec_job_id", String(255), nullable=False)
    submit_file = Column("submit_file", String(255), nullable=False)
    type_desc = Column(
        "type_desc",
        Enum(
            "unknown",
            "compute",
            "stage-in-tx",
            "stage-out-tx",
            "registration",
            "inter-site-tx",
            "create-dir",
            "staged-compute",
            "cleanup",
            "chmod",
            "dax",
            "dag",
            name="job_type_desc",
        ),
        nullable=False,
    )
    clustered = Column("clustered", Boolean, nullable=False)
    max_retries = Column("max_retries", Integer, nullable=False)
    executable = Column("executable", Text, nullable=False)
    argv = Column("argv", Text)
    task_count = Column("task_count", Integer, nullable=False)

    # Relationships
    parents = relation(
        lambda: Job,
        backref="children",
        cascade="all",
        secondary=lambda: JobEdge.__table__,
        primaryjoin=lambda: and_(
            Job.wf_id == JobEdge.wf_id,
            Job.exec_job_id == foreign(JobEdge.child_exec_job_id),
        ),
        secondaryjoin=lambda: and_(
            Job.wf_id == JobEdge.wf_id,
            Job.exec_job_id == foreign(JobEdge.parent_exec_job_id),
        ),
    )
    tasks = relation(
        lambda: Task, backref="job", cascade="all, delete-orphan", passive_deletes=True,
    )
    job_instances = relation(
        lambda: JobInstance,
        backref="job",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )


Job.__table_args__ = (
    UniqueConstraint(Job.wf_id, Job.exec_job_id, name="UNIQUE_JOB"),
    table_keywords,
)
Index("job_type_desc_COL", Job.type_desc)
Index("job_exec_job_id_COL", Job.exec_job_id)


class JobEdge(Base):
    """."""

    __tablename__ = "job_edge"
    __table_args__ = (table_keywords,)

    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        primary_key=True,
    )
    parent_exec_job_id = Column("parent_exec_job_id", String(255), primary_key=True)
    child_exec_job_id = Column("child_exec_job_id", String(255), primary_key=True)


class JobInstance(Base):
    """."""

    __tablename__ = "job_instance"

    job_instance_id = Column("job_instance_id", KeyInteger, primary_key=True)
    job_id = Column(
        "job_id",
        KeyInteger,
        ForeignKey(Job.job_id, ondelete="CASCADE"),
        nullable=False,
    )
    host_id = Column(
        "host_id", KeyInteger, ForeignKey(Host.host_id, ondelete="SET NULL"),
    )
    job_submit_seq = Column("job_submit_seq", Integer, nullable=False)
    sched_id = Column("sched_id", String(255))
    site = Column("site", String(255))
    user = Column("user", String(255))
    work_dir = Column("work_dir", Text)
    cluster_start = Column("cluster_start", TimestampType)
    cluster_duration = Column("cluster_duration", DurationType)
    local_duration = Column("local_duration", DurationType)
    subwf_id = Column(
        "subwf_id", KeyInteger, ForeignKey(Workflow.wf_id, ondelete="SET NULL"),
    )
    stdout_file = Column("stdout_file", String(255))
    stdout_text = Column("stdout_text", Text)
    stderr_file = Column("stderr_file", String(255))
    stderr_text = Column("stderr_text", Text)
    stdin_file = Column("stdin_file", String(255))
    multiplier_factor = Column("multiplier_factor", Integer, nullable=False, default=1)
    exitcode = Column("exitcode", Integer)

    # Relationships

    # PM-712 don't want merges to happen to invocation table .
    # setting lazy = false leads to a big join query when a job_instance is updated
    # with the postscript status.
    invocations = relation(
        lambda: Invocation,
        backref="job_instance",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    states = relation(
        lambda: Jobstate,
        backref="job_instance",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    sub_workflow = relation(
        lambda: Workflow,
        backref="job_instance",
        cascade="all, delete-orphan",
        single_parent=True,
    )
    host = relation(
        lambda: Host,
        backref="job_instance",
        cascade="all, delete-orphan",
        single_parent=True,
    )
    integrity_metrics = relation(
        lambda: IntegrityMetrics,
        backref="job_instance",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    tag = relation(
        lambda: Tag,
        backref="job_instance",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )


JobInstance.__table_args__ = (
    UniqueConstraint(
        JobInstance.job_id, JobInstance.job_submit_seq, name="UNIQUE_JOB_INSTANCE"
    ),
    table_keywords,
)

# Jobstate definition
# ==> Same information that currently goes into jobstate.log file,
#       obtained from dagman.out file
#
# job_id = from Job table (autogenerated)
# state = from dagman.out file (3rd column of jobstate.log file)
# timestamp = from dagman,out file (1st column of jobstate.log file)


class Jobstate(Base):
    """."""

    __tablename__ = "jobstate"
    __table_args__ = (table_keywords,)

    # All four columns are marked as primary key to produce the desired
    # effect - ie: it is the combo of the four columns that make a row
    # unique.
    job_instance_id = Column(
        "job_instance_id",
        KeyInteger,
        ForeignKey(JobInstance.job_instance_id, ondelete="CASCADE"),
        primary_key=True,
    )
    state = Column("state", String(255), primary_key=True)
    timestamp = Column(
        "timestamp", TimestampType, primary_key=True, default=time.time(),
    )
    jobstate_submit_seq = Column(
        "jobstate_submit_seq", Integer, nullable=False, primary_key=True
    )
    reason = Column("reason", Text)


Index("jobstate_jobstate_submit_seq_COL", Jobstate.jobstate_submit_seq)


class Tag(Base):
    """."""

    __tablename__ = "tag"

    tag_id = Column("tag_id", KeyInteger, primary_key=True)
    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        nullable=False,
    )
    job_instance_id = Column(
        "job_instance_id",
        KeyInteger,
        ForeignKey(JobInstance.job_instance_id, ondelete="CASCADE"),
        nullable=False,
    )
    name = Column("name", String(255), nullable=False)
    count = Column("count", Integer, nullable=False)


Tag.__table_args__ = (
    UniqueConstraint(Tag.wf_id, Tag.job_instance_id, name="UNIQUE_TAG"),
    table_keywords,
)


class Task(Base):
    """."""

    __tablename__ = "task"

    task_id = Column("task_id", KeyInteger, primary_key=True)
    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        nullable=False,
    )
    job_id = Column("job_id", KeyInteger, ForeignKey(Job.job_id, ondelete="SET NULL"),)
    abs_task_id = Column("abs_task_id", String(255), nullable=False)
    transformation = Column("transformation", Text, nullable=False)
    argv = Column("argv", Text)
    type_desc = Column("type_desc", String(255), nullable=False)

    # Relationships
    parents = relation(
        lambda: Task,
        backref="children",
        cascade="all",
        secondary=lambda: TaskEdge.__table__,
        primaryjoin=lambda: and_(
            Task.wf_id == TaskEdge.wf_id,
            Task.abs_task_id == foreign(TaskEdge.child_abs_task_id),
        ),
        secondaryjoin=lambda: and_(
            Task.wf_id == TaskEdge.wf_id,
            Task.abs_task_id == foreign(TaskEdge.parent_abs_task_id),
        ),
    )
    task_files = relation(
        lambda: WorkflowFiles,
        backref="task",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    files = relation(
        lambda: RCLFN,
        secondary=lambda: WorkflowFiles.__table__,
        primaryjoin=lambda: Task.task_id == WorkflowFiles.task_id,
        secondaryjoin=lambda: WorkflowFiles.lfn_id == RCLFN.lfn_id,
    )
    meta = relation(
        lambda: TaskMeta,
        backref="task",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )


Task.__table_args__ = (
    UniqueConstraint(Task.wf_id, Task.abs_task_id, name="UNIQUE_TASK"),
    table_keywords,
)
Index("task_abs_task_id_COL", Task.abs_task_id)
Index("task_wf_id_COL", Task.wf_id)


class TaskEdge(Base):
    """."""

    __tablename__ = "task_edge"
    __table_args__ = (table_keywords,)

    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        primary_key=True,
    )
    parent_abs_task_id = Column("parent_abs_task_id", String(255), primary_key=True)
    child_abs_task_id = Column("child_abs_task_id", String(255), primary_key=True)


class TaskMeta(Base):
    """."""

    __tablename__ = "task_meta"
    __table_args__ = (table_keywords,)

    task_id = Column(
        "task_id",
        KeyInteger,
        ForeignKey(Task.task_id, ondelete="CASCADE"),
        primary_key=True,
    )
    key = Column("key", String(255), primary_key=True)
    value = Column("value", String(255), nullable=False)


class Invocation(Base):
    """."""

    __tablename__ = "invocation"

    invocation_id = Column("invocation_id", KeyInteger, primary_key=True)
    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        nullable=False,
    )
    job_instance_id = Column(
        "job_instance_id",
        KeyInteger,
        ForeignKey(JobInstance.job_instance_id, ondelete="CASCADE"),
        nullable=False,
    )
    task_submit_seq = Column("task_submit_seq", Integer, nullable=False)
    start_time = Column(
        "start_time", TimestampType, nullable=False, default=time.time()
    )
    remote_duration = Column("remote_duration", DurationType, nullable=False)
    remote_cpu_time = Column("remote_cpu_time", DurationType)
    exitcode = Column("exitcode", Integer, nullable=False)
    transformation = Column("transformation", Text, nullable=False)
    executable = Column("executable", Text, nullable=False)
    argv = Column("argv", Text)
    abs_task_id = Column("abs_task_id", String(255))
    maxrss = Column("maxrss", Integer)
    avg_cpu = Column("avg_cpu", Numeric(precision=16, scale=6))


Invocation.__table_args__ = (
    UniqueConstraint(
        Invocation.job_instance_id, Invocation.task_submit_seq, name="UNIQUE_INVOCATION"
    ),
    table_keywords,
)
Index("invoc_abs_task_id_COL", Invocation.abs_task_id)
Index("invoc_wf_id_COL", Invocation.wf_id)


class IntegrityMetrics(Base):
    """."""

    __tablename__ = "integrity"

    integrity_id = Column("integrity_id", KeyInteger, primary_key=True)
    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        nullable=False,
    )
    job_instance_id = Column(
        "job_instance_id",
        KeyInteger,
        ForeignKey(JobInstance.job_instance_id, ondelete="CASCADE"),
        nullable=False,
    )
    type = Column(
        "type", Enum("check", "compute", name="integrity_type_desc"), nullable=False
    )
    file_type = Column(
        "file_type", Enum("input", "output", name="integrity_file_type_desc")
    )
    count = Column("count", Integer, nullable=False)
    duration = Column("duration", DurationType, nullable=False)


IntegrityMetrics.__table_args__ = (
    UniqueConstraint(
        IntegrityMetrics.job_instance_id,
        IntegrityMetrics.type,
        IntegrityMetrics.file_type,
        name="UNIQUE_INTEGRITY",
    ),
    table_keywords,
)


# ---------------------------------------------
# JDBCRC
# ---------------------------------------------


class RCLFN(Base):
    """."""

    __tablename__ = "rc_lfn"

    lfn_id = Column("lfn_id", KeyInteger, primary_key=True)
    lfn = Column("lfn", String(245), nullable=False)

    # Relationships
    pfns = relation(
        lambda: RCPFN,
        backref="lfn",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    meta = relation(
        lambda: RCMeta,
        backref="lfn",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )


RCLFN.__table_args__ = (
    UniqueConstraint(RCLFN.lfn, name="UNIQUE_LFN"),
    table_keywords,
)


class RCPFN(Base):
    """."""

    __tablename__ = "rc_pfn"

    pfn_id = Column("pfn_id", KeyInteger, primary_key=True)
    lfn_id = Column(
        "lfn_id",
        KeyInteger,
        ForeignKey(RCLFN.lfn_id, ondelete="CASCADE"),
        nullable=False,
    )
    pfn = Column("pfn", String(245), nullable=False)
    site = Column("site", String(245))


RCPFN.__table_args__ = (
    UniqueConstraint(RCPFN.lfn_id, RCPFN.pfn, RCPFN.site, name="UNIQUE_PFN"),
    table_keywords,
)


class RCMeta(Base):
    """."""

    __tablename__ = "rc_meta"
    __table_args__ = (table_keywords,)

    lfn_id = Column(
        "lfn_id",
        KeyInteger,
        ForeignKey(RCLFN.lfn_id, ondelete="CASCADE"),
        primary_key=True,
    )
    key = Column("key", String(245), primary_key=True)
    value = Column("value", String(245), nullable=False)


class WorkflowFiles(Base):
    """."""

    __tablename__ = "workflow_files"
    __table_args__ = (table_keywords,)

    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(Workflow.wf_id, ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    task_id = Column(
        "task_id",
        KeyInteger,
        ForeignKey(Task.task_id, ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    lfn_id = Column(
        "lfn_id",
        KeyInteger,
        ForeignKey(RCLFN.lfn_id, ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )
    file_type = Column("file_type", String(255))

    # Relationships
    lfn = relation(lambda: RCLFN, cascade="all, delete-orphan", single_parent=True)


# ---------------------------------------------
# DASHBOARD
# ---------------------------------------------


class MasterWorkflow(Base):
    """."""

    __tablename__ = "master_workflow"

    # ==> Information comes from braindump.txt file
    wf_id = Column("wf_id", KeyInteger, primary_key=True)
    wf_uuid = Column("wf_uuid", String(255), nullable=False)
    dax_label = Column("dax_label", String(255))
    dax_version = Column("dax_version", String(255))
    dax_file = Column("dax_file", String(255))
    dag_file_name = Column("dag_file_name", String(255))
    timestamp = Column("timestamp", TimestampType)
    submit_hostname = Column("submit_hostname", String(255))
    submit_dir = Column("submit_dir", Text)
    planner_arguments = Column("planner_arguments", Text)
    user = Column("user", String(255))
    grid_dn = Column("grid_dn", String(255))
    planner_version = Column("planner_version", String(255))
    db_url = Column("db_url", Text)
    archived = Column("archived", Boolean, nullable=False, default=0)

    # Relationships
    states = relation(
        lambda: MasterWorkflowstate,
        backref="workflow",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )


MasterWorkflow.__table_args__ = (
    UniqueConstraint(MasterWorkflow.wf_uuid, name="UNIQUE_MASTER_WF_UUID"),
    table_keywords,
)


class MasterWorkflowstate(Base):
    """."""

    __tablename__ = "master_workflowstate"
    __table_args__ = (table_keywords,)

    # All three columns are marked as primary key to produce the desired
    # effect - ie: it is the combo of the three columns that make a row
    # unique.
    wf_id = Column(
        "wf_id",
        KeyInteger,
        ForeignKey(MasterWorkflow.wf_id, ondelete="CASCADE"),
        primary_key=True,
    )
    state = Column(
        "state",
        Enum("WORKFLOW_STARTED", "WORKFLOW_TERMINATED", name="master_workflow_state"),
        primary_key=True,
    )
    timestamp = Column("timestamp", TimestampType, primary_key=True)
    restart_count = Column("restart_count", Integer, nullable=False)
    status = Column("status", Integer)
    reason = Column("reason", Text)


# ---------------------------------------------
# ENSEMBLE MANAGER
# ---------------------------------------------


class Ensemble(Base):
    """."""

    __tablename__ = "ensemble"

    id = Column("id", KeyInteger, primary_key=True)
    name = Column("name", String(100), nullable=False)
    created = Column("created", DateTime, nullable=False)
    updated = Column("updated", DateTime, nullable=False)
    state = Column(
        "state", Enum(*EnsembleStates, name="ensemble_state"), nullable=False
    )
    max_running = Column("max_running", Integer, nullable=False)
    max_planning = Column("max_planning", Integer, nullable=False)
    username = Column("username", String(100), nullable=False)


Ensemble.__table_args__ = (
    UniqueConstraint(Ensemble.name, Ensemble.username, name="UNIQUE_ENSEMBLE"),
    table_keywords,
)

# User mapper(..,...) to extend ensembles.Ensemble to the schema.Ensemble table
mapper(
    _Ensemble,
    Ensemble.__table__,
    properties={"workflows": relation(lambda: _EnsembleWorkflow)},
)


class EnsembleWorkflow(Base):
    """."""

    __tablename__ = "ensemble_workflow"

    id = Column("id", KeyInteger, primary_key=True)
    ensemble_id = Column(
        "ensemble_id", KeyInteger, ForeignKey(Ensemble.id), nullable=False
    )
    name = Column("name", String(100), nullable=False)
    basedir = Column("basedir", String(512), nullable=False)
    created = Column("created", DateTime, nullable=False)
    updated = Column("updated", DateTime, nullable=False)
    state = Column(
        "state",
        Enum(*EnsembleWorkflowStates, name="ensemble_wf_state"),
        nullable=False,
    )
    priority = Column("priority", Integer, nullable=False)
    wf_uuid = Column("wf_uuid", String(36))
    submitdir = Column("submitdir", String(512))
    plan_command = Column(
        "plan_command", String(1024), nullable=False, default="./plan.sh"
    )


EnsembleWorkflow.__table_args__ = (
    UniqueConstraint(
        EnsembleWorkflow.ensemble_id,
        EnsembleWorkflow.name,
        name="UNIQUE_ENSEMBLE_WORKFLOW",
    ),
    table_keywords,
)

mapper(
    _EnsembleWorkflow,
    EnsembleWorkflow.__table__,
    properties={"ensemble": relation(_Ensemble)},
)


class Trigger(Base):
    """."""

    __tablename__ = "trigger"
    _id = Column("id", KeyInteger, primary_key=True)
    ensemble_id = Column(
        "ensemble_id", KeyInteger, ForeignKey(Ensemble.id), nullable=False
    )
    name = Column("name", String(100), nullable=False)
    state = Column("state", Enum(*TriggerStates, name="trigger_state"), nullable=False)
    workflow = Column("workflow", Text(), nullable=False)
    args = Column("args", Text())
    _type = Column(
        "type",
        Enum(*(t.value for t in list(TriggerType)), name="trigger_type"),
        nullable=False,
    )


Trigger.__table_args__ = (
    UniqueConstraint(Trigger.ensemble_id, Trigger.name, name="UNIQUE_TRIGGER"),
    table_keywords,
)



"""
This file implements the Job class for pegasus-monitord.
"""

##
#  Copyright 2007-2012 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##

import collections
import json
import logging
import os
import re
from io import StringIO

from Pegasus.tools import utils

logger = logging.getLogger(__name__)

# Global variables
good_rsl = {"maxcputime": 1, "maxtime": 1, "maxwalltime": 1}
MAX_OUTPUT_LENGTH = 2 ** 16 - 1  # Only keep stdout to 64K

# some constants
NOOP_JOB_PREFIX = (
    "noop_"  # prefix for noop jobs for which .out and err files are not created
)

MONITORING_EVENT_START_MARKER = "@@@MONITORING_PAYLOAD - START@@@"
MONITORING_EVENT_END_MARKER = "@@@MONITORING_PAYLOAD - END@@@"

# Used in parse_sub_file
re_rsl_string = re.compile(r"^\s*globusrsl\W", re.IGNORECASE)
re_rsl_clean = re.compile(r"([-_])")
re_site_parse_gvds = re.compile(
    r"^\s*\+(pegasus|wf)_(site|resource)\s*=\s*([\'\"])?(\S+)\3"
)
re_parse_transformation = re.compile(r"^\s*\+pegasus_wf_xformation\s*=\s*(\S+)")
re_parse_derivation = re.compile(r"^\s*\+pegasus_wf_dax_job_id\s*=\s*(\S+)")
re_parse_multiplier_factor = re.compile(r"^\s*\+pegasus_cores\s=\s(\S+)")
re_parse_executable = re.compile(r"^\s*executable\s*=\s*(\S+)")
re_parse_arguments = re.compile(r'^\s*arguments\s*=\s*"([^"\r\n]*)"')
re_parse_environment = re.compile(r"^\s*environment\s*=\s*(.*)")
re_site_parse_euryale = re.compile(r"^\#!\s+site=(\S+)")
re_parse_property = re.compile(r"([^:= \t]+)\s*[:=]?\s*(.*)")
re_parse_input = re.compile(r"^\s*intput\s*=\s*(\S+)")
re_parse_output = re.compile(r"^\s*output\s*=\s*(\S+)")
re_parse_error = re.compile(r"^\s*error\s*=\s*(\S+)")
re_parse_job_class = re.compile(r"^\s*\+pegasus_job_class\s*=\s*(\S+)")
re_parse_pegasuslite_hostname = re.compile(
    r"^.*Executing on host\s*(\S+)(?:\s*IP=(\S+))*.*$", re.MULTILINE
)


TaskOutput = collections.namedtuple("TaskOutput", ["user_data", "events"])


class IntegrityMetric:
    """
    Class for storing integrity metrics and combining them based solely on type and file type combination
    """

    def __init__(self, type, file_type, count=0, succeeded=0, failed=0, duration=0.0):
        self.type = type
        self.file_type = file_type
        self.count = count
        self.succeeded = succeeded
        self.failed = failed
        self.duration = duration

    def __eq__(self, other):
        return self.type == other.type and self.file_type == other.file_type

    def __hash__(self):
        return hash(self.key())

    def __str__(self):
        return "({},{},{}, {}, {} , {})".format(
            self.type,
            self.file_type,
            self.count,
            self.succeeded,
            self.failed,
            self.duration,
        )

    def key(self):
        return self.type + ":" + self.file_type

    def merge(self, other):
        if self == other:
            self.count += other.count
            self.succeeded += other.succeeded
            self.failed += other.failed
            self.duration += other.duration
            return
        raise KeyError("Objects not compatible {} {}".format(self, other))


class Job:
    """
    Class used to keep information needed to track a particular job
    """

    JOBTYPE_TO_DESC = {
        "0": "unassigned",
        "1": "compute",
        "2": "stage-in-tx",
        "3": "stage-out-tx",
        "4": "registration",
        "5": "inter-site-tx",
        "6": "create-dir",
        "7": "stage_worker_tx",
        "8": "cleanup",
        "9": "chmod",
        "10": "dax",
        "11": "dag",
    }

    # Variables that describe a job, as per the Stampede schema
    # Some will be initialized in the init method, others will
    # get their values from the kickstart output file when a job
    # finished

    def __init__(self, wf_uuid, name, job_submit_dir, job_submit_seq):
        """
        This function initializes the job parameters with the
        information available when a job is detected in the
        "PRE_SCRIPT_STARTED" or the "SUBMIT" state. Other parameters
        will remain None until a job finishes and a kickstart output
        file can be parsed.
        """
        self._wf_uuid = wf_uuid
        self._exec_job_id = name
        self._job_submit_seq = job_submit_seq
        self._job_submit_dir = (
            job_submit_dir  # the submit directory all the job related files exist
        )
        self._sched_id = None
        self._site_name = None
        self._host_id = None
        self._host_ip = None
        self._remote_user = None
        self._remote_working_dir = None
        self._cluster_start_time = None
        self._cluster_duration = None
        self._job_type = None
        self._job_state = None
        self._job_state_seq = 0
        self._job_state_timestamp = None
        self._job_output_counter = 0
        self._pre_script_start = None
        self._pre_script_done = None
        self._pre_script_exitcode = None
        self._main_job_start = None
        self._main_job_done = None
        self._main_job_transformation = None
        self._main_job_derivation = None
        self._main_job_executable = None
        self._main_job_arguments = None
        self._main_job_exitcode = None
        self._main_job_multiplier_factor = None
        self._post_script_start = None
        self._post_script_done = None
        self._post_script_exitcode = None
        self._input_file = None
        self._output_file = None
        self._error_file = None
        self._stdout_text = None
        self._stderr_text = None
        self._additional_monitoring_events = []
        self._multipart_events = []
        self._cpu_attribs = None
        self._job_dagman_out = None  # _CONDOR_DAGMAN_LOG from environment
        # line for pegasus-plan and subdax_ jobs
        self._kickstart_parsed = False  # Flag indicating if the kickstart
        # output for this job was parsed or not
        self._has_rotated_stdout_err_files = (
            False  # Flag indicating whether we detected that job stdout|stderr
        )
        # was rotated or not, as is the default case.
        self._deferred_job_end_kwargs = None
        self._integrity_metrics = set()

    def _get_jobtype_desc(self):
        """
        Returns a textual description of the job type
        :return:
        """
        desc = "unknown"

        if self._job_type is None:
            return desc

        if self._job_type in Job.JOBTYPE_TO_DESC:
            desc = Job.JOBTYPE_TO_DESC[self._job_type]
        else:
            logger.error(
                "Unknown job type %s encountered for job %s"
                % (self._job_type, self._exec_job_id)
            )

        return desc

    def _add_multipart_events(self, events):
        """
        add multipart events to the job, separating any integrity metrics,
        since Integrity metrics are stored internally not as addditonal monitoring event
        :param events:
        :return:
        """
        for event in events:
            if not "multipart" in event:
                # Not this one... skip to the next
                logger.error(
                    " Mismatched multipart record %s in job %s"
                    % (event, self._exec_job_id)
                )
                continue
            if "integrity_summary" in event:
                # PM-1390 multipart events
                m = event["integrity_summary"]
                metric = IntegrityMetric(
                    type="check",  # For time being they always refer to verification
                    file_type="input",  # should be specified in multipart
                    succeeded=m["succeeded"] if "succeeded" in m else 0,
                    failed=m["failed"] if "failed" in m else 0,
                    duration=m["duration"] if "duration" in m else 0.0,
                )
                self.add_integrity_metric(metric)
            else:  # catch all
                self._multipart_events.append(event)

    def _add_additional_monitoring_events(self, events):
        """
        add monitoring events to the job, separating any integrity metrics,
        since Integrity metrics are stored internally not as addditonal monitoring event
        :param events:
        :return:
        """
        for event in events:
            if "monitoring_event" in event:
                # this is how integrity metrics were reported in 4.9.x series
                # as monitoring events; NOT multipart records in the job.out files
                name = event["monitoring_event"]
                if name == "int.metric":
                    # split elements in payload to IntegrityMetric
                    # add it internally for aggregation
                    for m in event["payload"]:
                        metric = IntegrityMetric(
                            type=m.get("event"),
                            file_type=m.get("file_type"),
                            count=m["count"] if "count" in m else 0,
                            succeeded=m["succeeded"] if "succeeded" in m else 0,
                            failed=m["failed"] if "failed" in m else 0,
                            duration=m["duration"] if "duration" in m else 0.0,
                        )
                        self.add_integrity_metric(metric)
                else:
                    self._additional_monitoring_events.append(event)
            else:  # catch all
                self._additional_monitoring_events.append(event)

    def add_integrity_metric(self, metric):
        """
        adds an integrity metric, if a metric with the same key already exists we retrive
        existing value and add the contents of metric passed
        :param metric:
        :return:
        """
        if metric is None:
            return

        for m in self._integrity_metrics:
            if metric == m:
                # add to existing metric
                m.merge(metric)
                break
        else:
            self._integrity_metrics.add(metric)

    def set_job_state(self, job_state, sched_id, timestamp, status):
        """
        This function sets the job state for this job. It also updates
        the times the main job and PRE/POST scripts start and finish.
        """
        self._job_state = job_state
        self._job_state_timestamp = int(timestamp)
        # Increment job state sequence
        self._job_state_seq = self._job_state_seq + 1

        # Set sched_id if we don't already have it...
        if self._sched_id is None:
            self._sched_id = sched_id
        # Record timestamp for certain job states
        if job_state == "PRE_SCRIPT_STARTED":
            self._pre_script_start = int(timestamp)
        elif job_state == "PRE_SCRIPT_SUCCESS" or job_state == "PRE_SCRIPT_FAILURE":
            self._pre_script_done = int(timestamp)
            self._pre_script_exitcode = utils.regular_to_raw(status)
        elif job_state == "POST_SCRIPT_STARTED":
            self._post_script_start = int(timestamp)
        elif job_state == "POST_SCRIPT_TERMINATED":
            self._post_script_done = int(timestamp)
        elif job_state == "EXECUTE":
            self._main_job_start = int(timestamp)
        elif job_state == "JOB_TERMINATED":
            self._main_job_done = int(timestamp)
        elif (
            job_state == "JOB_ABORTED"
            or job_state == "SUBMIT_FAILED"
            or job_state == "GLOBUS_SUBMIT_FAILED"
            or job_state == "GRID_SUBMIT_FAILED"
        ):
            self._main_job_done = int(
                timestamp
            )  # PM-805, PM-877 job was aborted or submit failed, good chance job terminated event did not happen.
        elif job_state == "JOB_SUCCESS" or job_state == "JOB_FAILURE":
            self._main_job_exitcode = utils.regular_to_raw(status)
        elif job_state == "POST_SCRIPT_SUCCESS" or job_state == "POST_SCRIPT_FAILURE":
            self._post_script_exitcode = utils.regular_to_raw(status)
            if self._main_job_done is None:
                # PM-1016 Missing JOB_TERMINATED event.
                self._main_job_done = int(timestamp)

    def parse_sub_file(self, stamp, submit_file):
        """
        This function parses a job's submit file and returns job
        planning information. In addition, we try to populate the job
        type from information in the submit file.
        # paramtr: stamp(IN): timestamp associated with the log line
        # paramtr: submit_file(IN): submit file name
        # globals: good_rsl(IN): which RSL keys constitute time requirements
        # returns: (largest job time requirement in minutes, destination site)
        # returns: (None, None) if sub file not found
        """
        parse_environment = False
        my_result = None
        my_site = None

        # Update stat record for submit file
        try:
            my_stats = os.stat(submit_file)
        except OSError:
            # Could not stat file
            logger.error("stat %s" % (submit_file))
            return my_result, my_site

        # Check submit file timestamp
        if stamp < my_stats[8]:  # mtime
            logger.info(
                "%s: sub file modified: job timestamp=%d, file mtime=%d, diff=%d"
                % (submit_file, stamp, my_stats[8], my_stats[8] - stamp)
            )

        # Check if we need to parse the environment line
        if self._exec_job_id.startswith("pegasus-plan") or self._exec_job_id.startswith(
            "subdax_"
        ):
            parse_environment = True

        try:
            SUB = open(submit_file)
        except OSError:
            logger.error("unable to parse %s" % (submit_file))
            return my_result, my_site

        # Parse submit file
        for my_line in SUB:
            if re_rsl_string.search(my_line):
                # Found RSL string, do parse now
                for my_match in re.findall(r"\(([^)]+)\)", my_line):
                    # Split into key and value
                    my_k, my_v = my_match.split("=", 1)
                    # Remove _- characters from string
                    my_k = re_rsl_clean.sub("", my_k)
                    if my_k.lower() in good_rsl and my_v > my_result:
                        try:
                            my_result = int(my_v)
                        except ValueError:
                            my_result = None
            elif re_site_parse_gvds.search(my_line):
                # GVDS agreement
                my_site = re_site_parse_gvds.search(my_line).group(4)
                self._site_name = my_site
            elif re_site_parse_euryale.search(my_line):
                # Euryale specific comment
                my_site = re_site_parse_euryale.search(my_line).group(1)
                self._site_name = my_site
            elif re_parse_transformation.search(my_line):
                # Found line with job transformation
                my_transformation = re_parse_transformation.search(my_line).group(1)
                # Remove quotes, if any
                my_transformation = my_transformation.strip('"')
                self._main_job_transformation = my_transformation
            elif re_parse_derivation.search(my_line):
                # Found line with job derivation
                my_derivation = re_parse_derivation.search(my_line).group(1)
                # Remove quotes, if any
                my_derivation = my_derivation.strip('"')
                if my_derivation == "null":
                    # If derivation is the "null" string, we don't want to take it
                    self._main_job_derivation = None
                else:
                    self._main_job_derivation = my_derivation
            elif re_parse_executable.search(my_line):
                # Found line with executable
                my_executable = re_parse_executable.search(my_line).group(1)
                # Remove quotes, if any
                my_executable = my_executable.strip('"')
                self._main_job_executable = my_executable
            elif re_parse_arguments.search(my_line):
                # Found line with arguments
                my_arguments = re_parse_arguments.search(my_line).group(1)
                # Remove quotes, if any
                my_arguments = my_arguments.strip('"')
                self._main_job_arguments = my_arguments
            elif re_parse_multiplier_factor.search(my_line):
                # Found line with multiplier_factor
                my_multiplier_factor = re_parse_multiplier_factor.search(my_line).group(
                    1
                )
                try:
                    self._main_job_multiplier_factor = int(my_multiplier_factor)
                except ValueError:
                    logger.warning(
                        "%s: cannot convert multiplier factor: %s"
                        % (os.path.basename(submit_file), my_multiplier_factor)
                    )
                    self._main_job_multiplier_factor = None
            elif re_parse_input.search(my_line):
                # Found line with input file
                my_input = re_parse_input.search(my_line).group(1)
                # Remove quotes, if any
                my_input = my_input.strip('"')
                self._input_file = os.path.normpath(my_input)
            elif re_parse_output.search(my_line):
                # Found line with output file
                my_output = re_parse_output.search(my_line).group(1)
                # Remove quotes, if any
                my_output = my_output.strip('"')
                self._output_file = os.path.normpath(my_output)
            elif re_parse_error.search(my_line):
                # Found line with error file
                my_error = re_parse_error.search(my_line).group(1)
                # Remove quotes, if any
                my_error = my_error.strip('"')
                self._error_file = os.path.normpath(my_error)
            elif parse_environment and re_parse_environment.search(my_line):
                self._job_dagman_out = self.extract_dagman_out_from_condor_env(my_line)
                if self._job_dagman_out is None:
                    logger.error(
                        "Unable to parse dagman out file from environment key %s in submit file for job %s"
                        % (my_line, self._exec_job_id)
                    )
            elif re_parse_job_class.search(my_line):
                self._job_type = re_parse_job_class.search(my_line).group(1)
        SUB.close()

        # All done!
        return my_result, my_site

    def extract_dagman_out_from_condor_env(self, condor_env):
        """
        This function extracts the dagman out file from the condor environment
        if one is specified

        :param condor_env: the environment line from the condor submit file
        :return: the dagman out file if detected else None
        """
        # Found line with environment
        env_value = re_parse_environment.search(condor_env).group(1)

        # strip any enclosing quotes if any
        stripped_env_value = re.sub(r'^"|"$', "", env_value)

        if len(env_value) == len(stripped_env_value):
            # we have old style condor environment with environment NOT ENCLOSED in double quotes
            # and split by ;
            sub_props = stripped_env_value.split(";")
        else:
            # we have new style condor environment with environment enclosed in double quotes
            # and split by whitespace
            sub_props = stripped_env_value.split(" ")

        dagman_out = None
        for sub_prop_line in sub_props:
            sub_prop_line = sub_prop_line.strip()  # Remove any spaces
            if len(sub_prop_line) == 0:
                continue
            sub_prop = re_parse_property.search(sub_prop_line)
            if sub_prop:
                if sub_prop.group(1) == "_CONDOR_DAGMAN_LOG":
                    dagman_out = sub_prop.group(2)
                    break

        return dagman_out

    def extract_job_info(self, kickstart_output):
        """
        This function reads the output from the kickstart parser and
        extracts the job information for the Stampede schema. It first
        looks for an invocation record, and then for a clustered
        record.

        Returns None if an error occurs, True if an invocation record
        was found, and False if it wasn't.
        """

        # Check if we have anything
        if len(kickstart_output) == 0:
            return None

        # Kickstart was parsed
        self._kickstart_parsed = True

        # PM-1157 we construct run dir from job submit dir
        self._job_submit_dir

        # Let's try to find an invocation record...
        my_invocation_found = False
        my_task_number = 0
        self._stdout_text = ""  # Initialize stdout
        stdout_text_list = []
        stdout_size = 0
        for my_record in kickstart_output:
            if "multipart" in my_record:
                # PM-1390 convert to integrity metrics
                logger.debug("Multipart record %s", my_record)
                self._add_multipart_events([my_record])
            elif not "invocation" in my_record:
                # Not this one... skip to the next
                logger.trace("Skipping %s", my_record)
                continue
            # Ok, we have an invocation record, extract the information we
            # need. Note that this may overwrite information obtained from
            # the submit file (e.g. the site_name).

            # Increment task_number
            my_task_number = my_task_number + 1

            if not my_invocation_found:
                # Things we only need to do once
                if "resource" in my_record:
                    self._site_name = my_record["resource"]
                if "user" in my_record:
                    self._remote_user = my_record["user"]
                if "cwd" in my_record:
                    self._remote_working_dir = my_record["cwd"]
                # We are done with this part
                my_invocation_found = True

            # PM-1488 for containers case we have to remap hostname for all invocation
            # records in a clustered job, not just the first one. otherwise the sqlite db
            # will have dangling host records in the host table that has docker container generated
            # names but no jobs associated with them.
            if "hostname" in my_record:
                ks_hostname = my_record["hostname"]
                if self._host_id is None:
                    # PM-1488 only set the hostname to kickstart reported one only if
                    # it is not determined already (PegasusLite case) by parsing the job err file
                    self._host_id = ks_hostname
                elif self._host_id != ks_hostname:
                    ks_hostaddr = my_record["hostaddr"]
                    if self._host_ip is not None:
                        # for 4.9 backward compatibilty where PegasusLite does not record IP
                        # we keep the kickstart reported ip address to allow for database
                        # population as host table requires an ip to be not null
                        my_record["hostaddr"] = self._host_ip

                    my_record["hostname"] = self._host_id
                    logger.trace(
                        "For job %s preferring %s %s over kickstart reported hostname %s %s"
                        % (
                            self._exec_job_id,
                            my_record["hostname"],
                            my_record["hostaddr"],
                            ks_hostname,
                            ks_hostaddr,
                        )
                    )

            # PM-1109 encode signal information if it exists
            signal_message = " "
            if "signalled" in my_record:
                # construct our own error message
                attrs = my_record["signalled"]
                signal_message = "Job was "
                if "action" in attrs:
                    signal_message += attrs["action"]
                if "signal" in attrs:
                    signal_message += " with signal " + attrs["signal"]

            # PM-641 optimization Modified string concatenation to a list join
            if "stdout" in my_record:
                task_output = self.split_task_output(my_record["stdout"])
                self._add_additional_monitoring_events(task_output.events)
                # PM-1152 we always attempt to store upto MAX_OUTPUT_LENGTH
                stdout = self.get_snippet_to_populate(
                    task_output.user_data, my_task_number, stdout_size, "stdout"
                )
                if stdout is not None:
                    try:
                        stdout_text_list.append(
                            utils.quote("#@ %d stdout\n" % (my_task_number))
                        )
                        stdout_text_list.append(utils.quote(stdout))
                        stdout_text_list.append(utils.quote("\n"))
                        stdout_size += len(stdout) + 20
                    except KeyError:
                        logger.exception(
                            "Unable to parse stdout section from kickstart record for task %s from file %s "
                            % (my_task_number, self.get_rotated_out_filename())
                        )

            if "stderr" in my_record:
                task_error = self.split_task_output(my_record["stderr"])
                # add the events to those retrieved from the application stderr
                self._add_additional_monitoring_events(task_error.events)
                # Note: we are populating task stderr from kickstart record to job stdout only
                stderr = self.get_snippet_to_populate(
                    signal_message + task_error.user_data,
                    my_task_number,
                    stdout_size,
                    "stderr",
                )
                if stderr is not None:
                    try:
                        stdout_text_list.append(
                            utils.quote("#@ %d stderr\n" % (my_task_number))
                        )
                        stdout_text_list.append(utils.quote(stderr))
                        stdout_text_list.append(utils.quote("\n"))
                        stdout_size += len(stderr) + 20
                    except KeyError:
                        logger.exception(
                            "Unable to parse stderr section from kickstart record for task %s from file %s "
                            % (my_task_number, self.get_rotated_out_filename())
                        )

            # PM-1398 pass cpu info
            if "cpu" in my_record:
                self._cpu_attribs = my_record["cpu"]

        if len(stdout_text_list) > 0:
            self._stdout_text = "".join(stdout_text_list)

            # PM-641 optimization merged encoding above
        # Now, we encode it!
        #        if self._stdout_text != "":
        #            self._stdout_text = utils.quote(self._stdout_text)

        if not my_invocation_found:
            logger.debug("cannot find invocation record in output")

        # Look for clustered record...
        my_cluster_found = False
        for my_record in kickstart_output:
            if not "clustered" in my_record:
                # Not this one... skip to the next
                continue
            # Ok found it, fill in cluster parameters
            if "duration" in my_record:
                self._cluster_duration = my_record["duration"]
            if "start" in my_record:
                # Convert timestamp to EPOCH
                my_start = utils.epochdate(my_record["start"])
                if my_start is not None:
                    self._cluster_start_time = my_start
            # No need to look further...
            my_cluster_found = True
            break

        if not my_cluster_found:
            logger.debug("cannot find cluster record in output")

        # Done populating Job class with information from the output file
        return my_invocation_found

    def get_job_state(self):
        """
        Returns the current job state
        :return:
        """
        return self._job_state

    def get_rotated_out_filename(self):
        """
        Returns the name of the rotated .out file for the job on the basis
        of the current counter
        """

        basename = self._output_file
        if self._has_rotated_stdout_err_files:
            basename += ".%03d" % (self._job_output_counter)

        return basename

    def get_rotated_err_filename(self):
        """
        Returns the name of the rotated .err file for the job on the basis
        of the current counter
        """

        basename = self._exec_job_id + ".err"
        if self._has_rotated_stdout_err_files:
            basename += ".%03d" % (self._job_output_counter)

        return basename

    def read_job_error_file(self, store_monitoring_events=True):
        """
        Reads the job error file and updates job structures to store the
        the stderr of the condor job and also attempts to parse the hostname
        from the stderr of the job

        :param store_monitoring_events: whether to store any parsed monitoring events in the job
        :return:
        """
        my_max_encoded_length = MAX_OUTPUT_LENGTH - 2000
        if self._error_file is None:
            # This is the case for SUBDAG jobs
            self._stderr_text = None
            return

        # Finally, read error file only
        run_dir = self._job_submit_dir
        basename = self.get_rotated_err_filename()
        my_err_file = os.path.join(run_dir, basename)

        try:
            ERR = open(my_err_file)
            # PM-1274 parse any monitoring events such as integrity related
            # from PegasusLite .err file
            job_stderr = self.split_task_output(ERR.read())
            buf = job_stderr.user_data

            if len(buf) > my_max_encoded_length:
                buf = buf[:my_max_encoded_length]
            self._stderr_text = utils.quote(buf)

            if store_monitoring_events:
                self._add_additional_monitoring_events(job_stderr.events)

            # PM-1355 attempt to determine the hostname from the pegasus lite job
            hostname_match = re_parse_pegasuslite_hostname.search(job_stderr.user_data)
            if hostname_match:
                # a match yes it is a PegasusLite job . gleam the hostname
                self._host_id = hostname_match.group(1)
                self._host_ip = hostname_match.group(2)

        except OSError:
            self._stderr_text = None
            if not self.is_noop_job():
                logger.warning(
                    "unable to read error file: %s, continuing..." % (my_err_file)
                )
        else:
            ERR.close()

    def read_job_out_file(self, out_file=None, store_monitoring_events=True):
        """
        This function reads both stdout and stderr files and populates
        these fields in the Job class.
        """
        my_max_encoded_length = MAX_OUTPUT_LENGTH - 2000
        if self._output_file is None:
            # This is the case for SUBDAG jobs
            self._stdout_text = None

        if out_file is None:
            # PM-1297 only construct relative path if out_file is not explicitly specified
            run_dir = self._job_submit_dir

            # PM-1157 output file has absolute path from submit file
            # interferes with replay mode on another directory
            # basename = self._output_file

            basename = self._exec_job_id + ".out"
            if self._has_rotated_stdout_err_files:
                basename += ".%03d" % (self._job_output_counter)
            out_file = os.path.join(run_dir, basename)

        try:
            OUT = open(out_file)
            job_stdout = self.split_task_output(OUT.read())
            buf = job_stdout.user_data
            if len(buf) > my_max_encoded_length:
                buf = buf[:my_max_encoded_length]
            self._stdout_text = utils.quote("#@ 1 stdout\n" + buf)

            if store_monitoring_events:
                self._add_additional_monitoring_events(job_stdout.events)
        except OSError:
            self._stdout_text = None
            if not self.is_noop_job():
                logger.warning(
                    "unable to read output file: %s, continuing..." % (out_file)
                )
        else:
            OUT.close()

    def is_noop_job(self):
        """
        A convenience method to indicate whether a job is a NOOP job or not
        :return:  True if a noop job else False
        """
        if self._exec_job_id is not None and self._exec_job_id.startswith(
            NOOP_JOB_PREFIX
        ):
            return True

        return False

    def get_snippet_to_populate(
        self, task_output, task_number, current_buffer_size, type
    ):
        """

        :param task_output:
        :param task number:          the task number
        :param current_buffer_size:  the current size of the buffer that is storing job stdout for all tasks
        :param type:                 whether stdout or stderr for logging
        :return:
        """
        # PM-1152 we always attempt to store upto MAX_OUTPUT_LENGTH
        # 20 is the rough estimate of number of extra characters added by URL encoding
        remaining = MAX_OUTPUT_LENGTH - current_buffer_size - 20
        task_output_size = len(task_output)
        stdout = None
        if task_output_size <= remaining:
            # we can store the whole stdout
            stdout = task_output
        else:
            logger.debug(
                "Only grabbing %s of %s for task %s from file %s "
                % (remaining, type, task_number, self.get_rotated_out_filename())
            )
            if remaining > 0:
                # we store only the first remaining chars
                stdout = task_output[: -(task_output_size - remaining)]

        return stdout

    def split_task_output(self, task_output):
        """
        Splits task output in to user app data and monitoring events for pegasus use
        :param task_output:
        :param type:                 whether stdout or stderr for logging
        :return:
        """

        events = []
        start = 0
        end = 0

        # print task_output
        start = task_output.find(MONITORING_EVENT_START_MARKER, start)

        if start == -1:
            # no monitoring marker found
            return TaskOutput(task_output, events)

        task_data = StringIO()
        try:
            while start != -1:
                task_data.write(task_output[end:start])
                end = task_output.find(MONITORING_EVENT_END_MARKER, start)
                payload = task_output[start + len(MONITORING_EVENT_START_MARKER) : end]
                try:
                    events.append(json.loads(payload))
                except Exception:
                    logger.error("Unable to convert payload %s to JSON" % payload)
                start = task_output.find(MONITORING_EVENT_START_MARKER, end)

            task_data.write(task_output[end + len(MONITORING_EVENT_END_MARKER) :])
        except Exception as e:
            logger.error(
                "Unable to parse monitoring events from job stdout for job %s"
                % self._exec_job_id
            )
            logger.exception(e)
            # return the whole task output as is
            return TaskOutput(task_data.getvalue(), events)

        return TaskOutput(task_data.getvalue(), events)

    def create_composite_job_event(self, job_inst_kwargs):
        """
        this creates a composite job event that also includes all information included in a job_inst.end event

        :param my_job:
        :param job_inst_kwargs:
        :return:
        """

        kwargs = {}

        # add on events associated to populate the job_instance table
        kwargs.update(job_inst_kwargs)

        # count any integrity errors
        error_count = 0
        for metric in self._integrity_metrics:
            error_count += metric.failed

        kwargs["int_error_count"] = error_count

        if self._host_id:
            kwargs["hostname"] = self._host_id

        job_type = self._get_jobtype_desc()
        kwargs["jobtype"] = job_type

        if self._cpu_attribs:
            for key in self._cpu_attribs:
                kwargs[key] = self._cpu_attribs[key]

        # PM-1398 for DIBBS we want task monitoring event that has metadata
        # to be included in the composite event also
        if self._additional_monitoring_events:
            for event in self._additional_monitoring_events:
                event_name = (
                    event["monitoring_event"]
                    if "monitoring_event" in event
                    else "monitoring.additional"
                )
                if event_name == "metadata":
                    # flatten out metadata key values into event
                    """
                    # sample event we want to be able to parse
                    {
                     "ts": 1437688574,
                     "monitoring_event": "metadata",
                     "payload": [
                       {
                         "name": "num_template_banks",
                         "value" : 3
                       },
                       {
                         "name": "event_name",
                         "value" : "binary start merger"
                       }
                      ]
                    }

                    """
                    payload = event["payload"] if "payload" in event else None
                    if payload is None:
                        logger.error("No payload retrieved from event %s" % event)
                    for m in event["payload"]:
                        if "name" in m and "value" in m:
                            kwargs["metadata__" + m["name"]] = m["value"]
                        else:
                            logger.error(
                                "Additional monitoring event of type metadata can only have name value pairs in payload %s"
                                % event
                            )

        # sanity check
        if job_type == "unknown" or job_type == "unassigned":
            logger.warning(
                "Job {} has unknown type {}".format(self._exec_job_id, job_type)
            )

        # if error_count > 0:
        #  print kwargs

        for event in self._multipart_events:
            # if len(event) != 1:
            #    # should be a dictionary with exactly one key and value
            #    logger.error("Malformed multipart event %s for job %s" %(event,self._exec_job_id))
            #    continue
            for key in event.keys():
                if key == "multipart":
                    continue
                kwargs[key] = event[key]

        return kwargs
